"""
Functions and objects dealing with color.
"""

import numpy as np

def h2r(_hex):
    """
    Convert a hex string to an RGB-tuple.
    """
    if _hex.startswith('#'):
        l = _hex[1:]
    else:
        l = _hex
    return list(bytes.fromhex(l))

def r2h(rgb):
    """
    Convert an RGB-tuple to a hex string.
    """
    return '#%02x%02x%02x' % tuple(rgb)

def tohex(color):
    """
    Convert any color to its hex string.
    """

    if type(color) == str:
        if len(color) in (6, 7):
            try:
                h2r(color)
                return color
            except:
                pass
        try:
            return hex_colors[color]
        except KeyError as e:
            raise ValueError("unknown color: '" + color +"'")
    elif type(color) in (list, tuple, np.ndarray) and len(color) == 3:
        return r2h(color)
    else:
        raise ValueError("Don't know how to interpret color " + str(color))

def torgb(color):
    """
    Convert any color to an rgb tuple.
    """

    if type(color) == str:
        if len(color) in (6, 7):
            try:
                return h2r(color)
            except:
                pass
        try:
            return colors[color]
        except KeyError as e:
            raise ValueError("unknown color: '" + str(color) +"'")
    elif type(color) in (list, tuple, np.ndarray) and len(color) == 3:
        return h2r(color)
    else:
        raise ValueError("Don't know how to interpret color " + str(color))


def brighter(rgb):
    """
    Make the color (rgb-tuple) a tad brighter.
    """
    _rgb = tuple([ int(np.sqrt(a/255) * 255) for a in rgb ])
    return _rgb

def darker(rgb):
    """
    Make the color (rgb-tuple) a tad darker.
    """
    _rgb = tuple([ int((a/255)**2 * 255) for a in rgb ])
    return _rgb

hex_colors = {
    # default
    'marine' : '#365663',
    'red' : '#ee6f51',
    'teal' : '#2a9d8f',
    'green' : '#83e377',
    'yellow' : '#e9c46a',
    'grey' : '#646464',
    'pink' : '#ff9de4',
    'blue' : '#52b7dc',
    'grape' : '#b75b9e',
    'light grape' : '#ff9de4',
    'light marine' : '#47829a',
    'light red' : '#ff9f92',
    'light teal' : '#45dbc8',
    'light green': '#a4ffbb',
    'light yellow': '#ffee9d',
    'light grey' : '#a1a1a1',
    'light blue' : '#b0dff0',
    'light pink' : '#ffcaf0',

    # french79
    'french79 marine' : '#555587',
    'french79 red' : '#e64b3f',
    'french79 blue' : '#94d9ce',
    'french79 yellow' : '#efc94b',
    'french79 green' : '#71b081',
    'french79 light marine' : '#47829a',
    'french79 light red' : '#ec9f80',
    'french79 light blue' : '#d6e4db',
    'french79 light yellow' : '#edd88d',
    'french79 light green' : '#a4ffbb',

    # brewer
    'brewer yellow': '#ffcf46',
    'brewer pink': '#ff62b2',
    'brewer teal': '#1fb78a',
    'brewer grape': '#8e88da',
    'brewer orange': '#ff9849',
    'brewer green': '#66a61e',
    'brewer grey': '#909090',
    'brewer brown': '#a6761d',
    'brewer marine': '#555587',
    'brewer light yellow': '#ffecb8',
    'brewer light pink': '#efadcf',
    'brewer light grape': '#c4c1ef',
    'brewer light teal': '#97ddc8',
    'brewer light orange': '#ffd2af',
    'brewer light green': '#ceeaaf',
    'brewer light grey': '#cecece',
    'brewer light brown': '#d3bb90',
    'brewer light marine' : '#47829a',
}

hex_bg_colors = {
    'dark' : '#253237',
    'light' : '#fff9e8',
    'french79': '#26263d',
    'brewer dark': '#2f2940',
    'brewer light': '#fff9e8',
    'dark pastel' : '#253237',
    'light pastel' : '#fff9e8',
    'french79 pastel': '#26263d',
    'brewer dark pastel': '#2f2940',
    'brewer light pastel': '#fff9e8',
}

hex_link_colors = {
    'dark': '#4b5a62', 
    'light': '#a1a1a1',
    'french79': '#4b5a62', 
    'brewer dark': '#4b5a62', 
    'brewer light': '#a1a1a1', 
    'dark pastel': '#4b5a62', 
    'light pastel': '#a1a1a1', 
    'french79 pastel': '#4b5a62', 
    'brewer dark pastel': '#4b5a62', 
    'brewer light pastel': '#a1a1a1', 
}

colors = { name: h2r(col) for name, col in hex_colors.items() }
bg_colors = { name: h2r(col) for name, col in hex_bg_colors.items() }
link_colors = { name: h2r(col) for name, col in hex_link_colors.items() }


accompanying_color = {}

# map accompanying color, i.e. if a color is light map to its "normal"
# color, if it is normal, map to its light variant
for color, _hex in colors.items():
    name = color.split(" ")
    colname = name[-1]
    if len(name) == 1:
        accompanying_color[color] = 'light '+colname
    elif len(name) == 2:
        if name[0] == 'light':
            accompanying_color[color] = colname
        else:
            accompanying_color[color] = name[0] + ' light ' + colname
    elif len(name) == 3:
        accompanying_color[color] = name[0] + ' ' + colname

#for a, b in accompanying_color.items():
#    print(a,"  ----  ",b)

dark = [
            'marine',
            'red',
            'teal',
            'green',
            'yellow',
            'grey',
            'grape',
            'light marine',
            'light red',
            'light teal',
            'light green',
            'light yellow',
            'light grey',
            'light grape',
    ]

dark_pastel = [
            'light marine',
            'light red',
            'light teal',
            'light green',
            'light yellow',
            'light grey',
            'light grape',
            'marine',
            'red',
            'teal',
            'green',
            'yellow',
            'grey',
            'grape',
        ]

light = [
            'light yellow',
            'red',
            'teal',
            'green',
            'blue',
            'grey',
            'pink',
            'yellow',
            'light red',
            'light teal',
            'light green',
            'light blue',
            'light grey',
            'light pink',
        ]

light_pastel = [
            'yellow',
            'light red',
            'light teal',
            'light green',
            'light blue',
            'light grey',
            'light pink',
            'light yellow',
            'red',
            'teal',
            'green',
            'blue',
            'grey',
            'pink',
        ]

french79 = [
            'french79 marine',
            'french79 red',
            'french79 blue',
            'french79 yellow',
            'french79 green',
            'grey',
            'grape',
            'french79 light marine',
            'french79 light red',
            'french79 light blue',
            'french79 light yellow',
            'french79 light green',
            'light grey',
            'light grape',
        ]


french79_pastel = [
            'french79 light marine',
            'french79 light red',
            'french79 light blue',
            'french79 light yellow',
            'french79 light green',
            'light grey',
            'light grape',
            'french79 marine',
            'french79 red',
            'french79 blue',
            'french79 yellow',
            'french79 green',
            'grey',
            'grape',
        ]

brewer_light = [
            'brewer light yellow',
            'brewer pink',
            'brewer teal',
            'brewer grape',
            'brewer orange',
            'brewer green',
            'brewer grey',
            'brewer brown',

            'brewer yellow',
            'brewer light pink',
            'brewer light teal',
            'brewer light grape',
            'brewer light orange',
            'brewer light green',
            'brewer light grey',
            'brewer light brown',
        ]

brewer_light_pastel = [
            'brewer yellow',
            'brewer light pink',
            'brewer light teal',
            'brewer light grape',
            'brewer light orange',
            'brewer light green',
            'brewer light grey',
            'brewer light brown',

            'brewer light yellow',
            'brewer pink',
            'brewer teal',
            'brewer grape',
            'brewer orange',
            'brewer green',
            'brewer grey',
            'brewer brown',
        ]

brewer_dark = [
            'brewer marine',
            'brewer pink',
            'brewer teal',
            'brewer yellow',
            'brewer green',
            'brewer orange',
            'brewer grey',
            'brewer brown',

            'brewer light marine',
            'brewer light pink',
            'brewer light teal',
            'brewer light yellow',
            'brewer light green',
            'brewer light orange',
            'brewer light grey',
            'brewer light brown',
        ]

brewer_dark_pastel = [
            'brewer light marine',
            'brewer light pink',
            'brewer light teal',
            'brewer light yellow',
            'brewer light green',
            'brewer light orange',
            'brewer light grey',
            'brewer light brown',

            'brewer marine',
            'brewer pink',
            'brewer teal',
            'brewer yellow',
            'brewer green',
            'brewer orange',
            'brewer grey',
            'brewer brown',
        ]

palettes = {
       'dark':                  dark,
       'dark pastel':           dark_pastel,
       'light':                 light,
       'light pastel':          light_pastel,
       'french79':              french79,
       'french79 pastel':       french79_pastel,
       'brewer light':          brewer_light,
       'brewer light pastel':   brewer_light_pastel,
       'brewer dark':           brewer_dark,
       'brewer dark pastel':    brewer_dark_pastel,
   }

if __name__ == "__main__": # pragma : no cover

    from matplotlib.collections import LineCollection, EllipseCollection
    import matplotlib as mpl
    mpl.rcParams['font.sans-serif'] = ['Helvetica','Arial']
    import matplotlib.pyplot as pl

    for palette, this_palette in palettes.items():
        #if palette.endswith(' pastel'):
        #    continue

        fig, ax = pl.subplots(1,1)
        N = len(palette)
        firstcol = this_palette[:len(this_palette)//2]
        secondcol = this_palette[len(this_palette)//2:]
        islight = 'light' in palette
        bgcolor = hex_bg_colors[palette]
        print(firstcol, secondcol)
        if islight:
            fontcolor = 'k'
        else:
            fontcolor = 'w'

        r = 0.125
        dy = 0.5
        dx = 0.2
        XY = []
        size = []
        node_colors = []
        
        for i in range(len(firstcol)):
            firstname = firstcol[i] + ' // ' + hex_colors[firstcol[i]][1:].upper()
            secondname = hex_colors[secondcol[i]][1:].upper() + ' // ' + secondcol[i]
            XY.append([-dx, -i*dy-3])
            XY.append([+dx, -i*dy-3])
            size.extend([2*r, 2*r])
            node_colors.append(hex_colors[firstcol[i]])
            node_colors.append(hex_colors[secondcol[i]])

            ax.text(-dx-2*r, -i*dy-3, firstname, transform=ax.transData, ha='right',va='center', color=fontcolor)
            ax.text(+dx+2*r, -i*dy-3, secondname, transform=ax.transData, ha='left',va='center', color=fontcolor)

        circles = EllipseCollection(
            size, size, np.zeros_like(size),
            offsets=np.array(XY),
            units='x',
            transOffset=ax.transData,
            facecolors=node_colors,
            linewidths=0.5,
            edgecolors='k',
        )
        
        ax.add_collection(circles)
        ax.set_xlim([-3,3])
        ax.set_ylim([-8*dy-3,-2])
        ax.axis('off')
        ax.set_facecolor(bgcolor)
        fig.patch.set_facecolor(bgcolor)
        ax.set_title(palette.upper(),color=fontcolor)
        print(palette)

    pl.show()




"""
Some helper plot functions based on matplotlib.
These are just some code snippets that I use
regularly and not supposed to be of great merit otherwise.

Install matplotlib and bfmplot to use this module

.. code::

    matplotlib>=3.0.0
    bfmplot>=0.1.0
"""

import matplotlib.pyplot as pl
import matplotlib as mpl

import numpy as np

from epipack.colors import hex_colors, palettes

colors = [hex_colors[c] for c in  palettes['dark']]
mpl.rcParams['axes.prop_cycle'] = mpl.cycler(color=colors)

def strip_axis(ax,horizontal='right'):
    """Remove the right and the top axis"""
    if horizontal == 'right':
        anti_horizontal = 'left'
    else:
        anti_horizontal = 'right'
        ax.yaxis.set_label_position("right")

    ax.spines[horizontal].set_visible(False)
    ax.spines['top'].set_visible(False)

    ax.yaxis.set_ticks_position(anti_horizontal)
    ax.xaxis.set_ticks_position('bottom')

def plot(t,result,ax=None, curve_label_format='{}',figsize=None):
    """
    Plot an epipack result.

    Parameters
    ==========
    t : numpy.ndarray
        Sampling times.
    result : dict
        Mapping compartments to incidence time series.
    ax : matplotlib.axis.Axis, default = None
        The axis on which to plot
    curve_label_format : str, default = '{}'
        How to display a curve label
    figsize : tuple, default = None
        A tuple containing width and height of the figure
        that's produced

    Returns
    =======
    ax : matplotlib.axis.Axis
        The axis on which results were drawn
    """
    if ax is None:
        fig, ax = pl.subplots(1,1,figsize=figsize)
    _res = list(result.values())[0]
    N = np.zeros_like(_res)
    for C, timeseries in result.items():
        ax.plot(t, timeseries, label=curve_label_format.format(str(C)))
        N += timeseries
    strip_axis(ax)
    ax.set_xlim([t.min(), t.max()])
    ax.set_ylim([0,N.max()])
    ax.set_xlabel('time')
    ax.set_ylabel('frequency')
    ax.get_figure().tight_layout()
    return ax



"""
Provides an API to define  epidemiological models.
"""

import numpy as np
import scipy.sparse as sprs

import warnings

from epipack.integrators import (
        IntegrationMixin,
        time_leap_newton,
        time_leap_ivp,
    )

from epipack.process_conversions import (
        processes_to_events,
        transition_processes_to_events,
        fission_processes_to_events,
        fusion_processes_to_events,
        transmission_processes_to_events,
    )

from scipy.optimize import newton
from scipy.integrate import quad

def custom_choice(p):
    """
    Return an index of normalized probability
    vector ``p`` with probability equal to
    the corresponding entry in ``p``.
    """
    return np.argmin(np.cumsum(p)<np.random.rand())

class ConstantBirthRate():
    """
    Will be used as a function of
    time ``t`` and state ``y``,
    returning a rate value.

    Parameters
    ----------
    rate : float
        Constant rate value
    """
    def __init__(self, rate):
        self.rate = rate

    def __call__(self, t, y):
        return self.rate

class DynamicBirthRate():
    """
    Will be used as a function of
    time ``t`` and state ``y``,
    returning a rate value.

    Parameters
    ----------
    rate : function
        Function of time ``t`` and state ``y``
    """

    def __init__(self, rate):
        self.rate = rate

    def __call__(self, t, y):
        return self.rate(t,y)

class ConstantLinearRate:
    """
    Will be used as a function of
    time ``t`` and state ``y``,
    returning a rate value.

    Parameters
    ----------
    rate : float
        Constant rate value
    comp0 : int
        Index of the corresponding reacting
        component. The incidence of this component
        will be multiplied with the
        value of ``rate``.
    """

    def __init__(self, rate, comp0):
        self.rate = rate
        self.comp0 = comp0

    def __call__(self, t, y):
        return self.rate * y[self.comp0]

class DynamicLinearRate:
    """
    Will be used as a function of
    time ``t`` and state ``y``,
    returning a rate value.

    Parameters
    ----------
    rate : function
        Function of time ``t`` and state ``y``
    comp0 : int
        Index of the corresponding reacting
        component. The incidence of this component
        will be multiplied with the
        value of ``rate``.
    """

    def __init__(self, rate, comp0):
        self.rate = rate
        self.comp0 = comp0

    def __call__(self, t, y):
        return self.rate(t,y) * y[self.comp0]

class ConstantQuadraticRate:
    """
    Will be used as a function of
    time ``t`` and state ``y``,
    returning a rate value.

    Parameters
    ----------
    rate : float
        Constant rate value
    comp0 : int
        Index of one of the reacting
        components. The incidence of this component
        will be multiplied with the
        value of ``rate``.
    comp1 : int
        Index of the other reacting
        component. The incidence of this component
        will be multiplied with the
        value of ``rate``.
    """

    def __init__(self, rate, comp0, comp1):
        self.rate = rate
        self.comp0 = comp0
        self.comp1 = comp1

    def __call__(self, t, y):
        return self.rate * y[self.comp0] * y[self.comp1]

class DynamicQuadraticRate:
    """
    Will be used as a function of
    time ``t`` and state ``y``,
    returning a rate value.

    Parameters
    ----------
    rate : function
        Function of time ``t`` and state ``y``
    comp0 : int
        Index of one of the reacting
        components. The incidence of this component
        will be multiplied with the
        value of ``rate``.
    comp1 : int
        Index of the other reacting
        component. The incidence of this component
        will be multiplied with the
        value of ``rate``.
    """

    def __init__(self, rate, comp0, comp1):
        self.rate = rate
        self.comp0 = comp0
        self.comp1 = comp1

    def __call__(self, t, y):
        return self.rate(t,y) * y[self.comp0] * y[self.comp1]

class EpiModel(IntegrationMixin):
    """
    A general class to define a standard
    mean-field compartmental
    epidemiological model, based on reaction
    events.

    Parameters
    ----------
    compartments : :obj:`list` of :obj:`string`
        A list containing compartment strings.
    initial_population_size : float, default = 1.0
        The population size at :math:`t = 0`.
    correct_for_dynamical_population_size : bool, default = False
        If ``True``, the quadratic coupling terms will be
        divided by the population size.
    integral_solver : str, default = 'solve_ivp'
        Whether or not to use the initial-value solver ``solve_ivp``.
        to determine a time leap for time-varying rates.
        If not ``'solve_ivp'``, a Newton-Raphson method will be
        used on the upper bound of a quad-integrator.

    Attributes
    ----------
    compartments : list
        A list containing strings or hashable types that describe each compartment,
        (e.g. "S", "I", etc.).
    compartment_ids: dict
        Maps compartments to their indices in ``self.compartments``.
    N_comp : :obj:`int`
        Number of compartments (including population number)
    initial_population_size : float
        The population size at :math:`t = 0`.
    correct_for_dynamical_population_size : bool
        If ``True``, the quadratic coupling terms will be
        divided by the sum of all compartments, otherwise they
        will be divided by the initial population size.
    birth_rate_functions : list of ConstantBirthRate or DynamicBirthRate
        A list of functions that return rate values based on time ``t``
        and state vector ``y``. Each entry corresponds to an event update
        in ``self.birth_event_updates``.
    birth_event_updates : list of numpy.ndarray
        A list of vectors. Each entry corresponds to a rate in
        ``birth_rate_functions`` and quantifies the change in
        individual counts in the compartments.
    linear_rate_functions : list of ConstantLinearRate or DynamicLinearRate
        A list of functions that return rate values based on time ``t``
        and state vector ``y``. Each entry corresponds to an event update
        in ``self.linear_event_updates``.
    linear_event_updates : list of numpy.ndarray
        A list of vectors. Each entry corresponds to a rate in
        ``linear_rate_functions`` and quantifies the change in
        individual counts in the compartments.
    quadratic_rate_functions : list of ConstantQuadraticRate or DynamicQuadraticRate
        A list of functions that return rate values based on time ``t``
        and state vector ``y``. Each entry corresponds to an event update
        in ``self.quadratic_event_updates``.
    quadratic_event_updates : list of numpy.ndarray
        A list of vectors. Each entry corresponds to a rate in
        ``quadratic_rate_functions`` and quantifies the change in
        individual counts in the compartments.
    y0 : numpy.ndarray
        The initial conditions.
    rates_have_explicit_time_dependence : bool
        Internal switch that's flipped when a non-constant
        rate is passed to the model.
    use_ivp_solver : bool
        Whether or not to use the initial-value solver
        to determine a time leap for time-varying rates.
        If ``False``, a Newton-Raphson method will be
        used on the upper bound of a quad-integrator.

    Example
    -------

    .. code:: python

        >>> epi = EpiModel(["S","I","R"])
        >>> print(epi.compartments)
        [ "S", "I", "R" ]


    """

    def __init__(self,
                      compartments,
                      initial_population_size=1,
                      correct_for_dynamical_population_size=False,
                      integral_solver='solve_ivp',
                  ):

        self.y0 = None

        self.compartments = list(compartments)
        self.compartment_ids = { C: iC for iC, C in enumerate(compartments) }
        self.N_comp = len(self.compartments)

        self.initial_population_size = initial_population_size
        self.correct_for_dynamical_population_size = correct_for_dynamical_population_size

        self.birth_rate_functions = []
        self.birth_event_updates = []
        self.linear_rate_functions = []
        self.linear_event_updates = []
        self.quadratic_rate_functions = []
        self.quadratic_event_updates = []

        self.birth_events = []
        self.linear_events = []
        self.quadratic_events = []

        self.use_ivp_solver = integral_solver == 'solve_ivp'

        self.rates_have_explicit_time_dependence = False

    def get_compartment_id(self,C):
        """Get the integer ID of a compartment ``C``"""
        return self.compartment_ids[C]

    def get_compartment(self,iC):
        """Get the compartment, given an integer ID ``iC``"""
        return self.compartments[iC]

    def set_processes(self,
                      process_list,
                      allow_nonzero_column_sums=False,
                      reset_events=True,
                      ignore_rate_position_checks=False):
        """
        Converts a list of reaction process tuples to event tuples and sets the rates for this model.

        Parameters
        ----------
        process_list : :obj:`list` of :obj:`tuple`
            A list containing reaction processes in terms of tuples.

            .. code:: python

                [
                    # transition process
                    ( source_compartment, rate, target_compartment),

                    # transmission process
                    ( coupling_compartment_0, coupling_compartment_1, rate, target_compartment_0, target_ccompartment_1),

                    # fission process
                    ( source_compartment, rate, target_compartment_0, target_ccompartment_1),

                    # fusion process
                    ( source_compartment_0, source_compartment_1, rate, target_compartment),

                    # death process
                    ( source_compartment, rate, None),

                    # birth process
                    ( None, rate, target_compartment),
                ]
        allow_nonzero_column_sums : bool, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.
        reset_events : bool, default : True
            If this is `True`, reset all events to zero before setting the new ones.
        ignore_rate_position_checks : bool, default = False
            This function usually checks whether the rate of
            a reaction is positioned correctly. You can
            turn this behavior off for transition, birth, death, and
            transmission processes. (Useful if you want to define
            symbolic transmission processes that are compartment-dependent).
        """

        quadratic_events, linear_events = processes_to_events(process_list,
                                                              self.compartments,
                                                              ignore_rate_position_checks)
        self.set_linear_events(linear_events,
                               allow_nonzero_column_sums=allow_nonzero_column_sums,
                               reset_events=reset_events,
                               )
        self.set_quadratic_events(quadratic_events,
                                  allow_nonzero_column_sums=allow_nonzero_column_sums,
                                  reset_events=reset_events,
                                  )

        return self

    def _rate_has_functional_dependency(self,rate):
        if callable(rate):
            t = [0,1,2,10000,-10000]
            y = np.ones(self.N_comp)
            test = np.array([ rate(_t, y) for _t in t ])
            has_time_dependence = not np.all(test == test[0])
            self.rates_have_explicit_time_dependence = \
                    self.rates_have_explicit_time_dependence or has_time_dependence
            return True
        else:
            return False

    def set_linear_events(self,
                          event_list,
                          allow_nonzero_column_sums=False,
                          reset_events=True):
        r"""
        Define the linear transition events between compartments.

        Parameters
        ==========
        event_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transition events in the
            following format:

            .. code:: python

                [
                    (
                        ("affected_compartment_0",),
                        rate,
                        [
                            ("affected_compartment_0", dN0),
                            ("affected_compartment_1", dN1),
                            ...
                        ],
                     ),
                    ...
                ]

        allow_nonzero_column_sums : :obj:`bool`, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.
        reset_events : bool, default : True
            Whether to reset all linear events to zero before
            converting those.

        Example
        -------
        For an SEIR model with infectious period ``tau``
        and incubation period ``theta``.

        .. code:: python

            epi.set_linear_events([
                ( ("E",),
                  1/theta,
                  [ ("E", -1), ("I", +1) ]
                ),
                ( ("I",),
                  1/tau,
                  [ ("I", -1), ("R", +1) ]
                ),
            ])

        Read as "compartment E reacts with rate :math:`1/\theta`
        which leads to the decay of one E particle to one I particle."
        """

        if reset_events:
            birth_rate_functions = []
            birth_event_updates = []
            linear_rate_functions = []
            linear_event_updates = []
            birth_events = []
            linear_events = []
        else:
            linear_event_updates = list(self.linear_event_updates)
            birth_event_updates = list(self.birth_event_updates)
            linear_rate_functions = list(self.linear_rate_functions)
            birth_rate_functions = list(self.birth_rate_functions)
            birth_events = list(self.birth_events)
            linear_events = list(self.linear_events)


        for acting_compartments, rate, affected_compartments in event_list:

            dy = np.zeros(self.N_comp)
            for trg, change in affected_compartments:
                _t = self.get_compartment_id(trg)
                dy[_t] += change

            if acting_compartments[0] is None:
                if self._rate_has_functional_dependency(rate):
                    this_rate = DynamicBirthRate(rate)
                else:
                    this_rate = ConstantBirthRate(rate)
                birth_event_updates.append( dy )
                birth_rate_functions.append( this_rate )
                birth_events.append((acting_compartments, rate, affected_compartments))
            else:
                _s = self.get_compartment_id(acting_compartments[0])
                if self._rate_has_functional_dependency(rate):
                    this_rate = DynamicLinearRate(rate, _s)
                else:
                    this_rate = ConstantLinearRate(rate, _s)
                linear_event_updates.append( dy )
                linear_rate_functions.append( this_rate )
                linear_events.append((acting_compartments, rate, affected_compartments))

            if dy.sum() != 0 and not self.correct_for_dynamical_population_size:
                warnings.warn("This model has processes with a fluctuating "+\
                        "number of agents. Consider correcting the rates dynamically with "+\
                        "the attribute correct_for_dynamical_population_size = True")


        if not allow_nonzero_column_sums and len(linear_rate_functions)>0:
            _y = np.ones(self.N_comp)
            test = sum([r(0,_y) * dy for dy, r in zip (linear_event_updates, linear_rate_functions)])
            test += sum([r(0,_y) * dy for dy, r in zip (birth_event_updates, birth_rate_functions)])
            test_sum = test.sum()
            if np.abs(test_sum) > 1e-15:
                warnings.warn("events do not sum to zero for each column:" + str(test_sum))

        self.linear_event_updates = linear_event_updates
        self.linear_rate_functions = linear_rate_functions
        self.birth_event_updates = birth_event_updates
        self.birth_rate_functions = birth_rate_functions
        self.linear_events = linear_events
        self.birth_events = birth_events

        return self


    def add_transition_processes(self,process_list):
        """
        Define the linear transition processes between compartments.

        Parameters
        ==========
        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions events in the following format:

            .. code:: python

                [
                    ( source_compartment, rate, target_compartment ),
                    ...
                ]

        Example
        =======

        For an SEIR model.

        .. code:: python

            epi.add_transition_processes([
                ("E", symptomatic_rate, "I" ),
                ("I", recovery_rate, "R" ),
            ])

        """

        linear_events = transition_processes_to_events(process_list)

        return self.set_linear_events(linear_events,
                                      reset_events=False,
                                      allow_nonzero_column_sums=True)

    def add_fission_processes(self,process_list):
        """
        Define linear fission processes between compartments.

        Parameters
        ==========

        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains fission rates in the following format:

            .. code:: python

                [
                    ("source_compartment", rate, "target_compartment_0", "target_compartment_1" ),
                    ...
                ]

        Example
        -------

        For pure exponential growth of compartment `B`.

        .. code:: python

            epi.add_fission_processes([
                ("B", growth_event, "B", "B" ),
            ])

        """
        linear_events = fission_processes_to_events(process_list)

        return self.set_linear_events(linear_events,
                                      reset_events=False,
                                      allow_nonzero_column_sums=True)

    def add_fusion_processes(self,process_list):
        """
        Define fusion processes between compartments.

        Parameters
        ==========

        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains fission rates in the following format:

            .. code:: python

                [
                    ("coupling_compartment_0", "coupling_compartment_1", rate, "target_compartment_0" ),
                    ...
                ]

        Example
        -------

        Fusion of reactants "A", and "B" to form "C".

        .. code:: python

            epi.add_fusion_processes([
                ("A", "B", reaction_rate, "C" ),
            ])

        """
        quadratic_events = fusion_processes_to_events(process_list)

        return self.set_quadratic_events(quadratic_events,
                                         reset_events=False, 
                                         allow_nonzero_column_sums=True)

    def add_transmission_processes(self,process_list):
        r"""
        A wrapper to define quadratic process rates
        through transmission reaction equations.
        Note that in stochastic network/agent simulations, the transmission
        rate is equal to a rate per link. For the mean-field ODEs,
        the rates provided to this function will just be equal 
        to the prefactor of the respective quadratic terms.

        For instance, if you analyze an SIR system and simulate
        on a network of mean degree :math:`k_0`,
        a basic reproduction number :math:`R_0`, and a 
        recovery rate :math:`\mu`, you would define the single 
        link transmission process as 

            .. code:: python

                ("I", "S", R_0/k_0 * mu, "I", "I")

        For the mean-field system here, the corresponding reaction equation would read

            .. code:: python

                ("I", "S", R_0 * mu, "I", "I")

        Parameters
        ----------
        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ("source_compartment", 
                     "target_compartment_initial",
                     rate 
                     "source_compartment", 
                     "target_compartment_final", 
                     ),
                    ...
                ]

        Example
        -------

        For an SEIR model.

        .. code:: python

            epi.add_transmission_processes([
                ("I", "S", +1, "I", "E" ),
            ])

        """
        quadratic_events = transmission_processes_to_events(process_list)

        return self.set_quadratic_events(quadratic_events,
                                         reset_events=False, 
                                         allow_nonzero_column_sums=True)

    def add_quadratic_events(self,
                             event_list,
                             allow_nonzero_column_sums=False):
        """
        Add quadratic events without resetting the existing event terms.
        See :func:`epipack.numeric_epi_models.EpiModel.set_quadratic_events` for docstring.
        """

        return self.set_quadratic_events(event_list,
                                         reset_events=False,
                                         allow_nonzero_column_sums=allow_nonzero_column_sums,
                                         )

    def add_linear_events(self,
                          event_list,
                          allow_nonzero_column_sums=False):
        """
        Add linear events without resetting the existing event terms.
        See :func:`epipack.numeric_epi_models.EpiModel.set_linear_events` for docstring.
        """

        return self.set_linear_events(event_list,
                                      reset_events=False,
                                      allow_nonzero_column_sums=allow_nonzero_column_sums
                                      )

    def set_quadratic_events(self,
                             event_list,
                             allow_nonzero_column_sums=False,
                             reset_events=True,
                             ):
        r"""
        Define quadratic transition events between compartments.

        Parameters
        ----------
        event_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transmission events in the following format:

            .. code:: python

                [
                    (
                        ("coupling_compartment_0", "coupling_compartment_1"),
                        rate,
                        [
                            ("affected_compartment_0", dN0),
                            ("affected_compartment_1", dN1),
                            ...
                        ],
                     ),
                    ...
                ]
        allow_nonzero_column_sums : :obj:`bool`, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.
        reset_events : bool, default : True
            Whether to reset all linear events to zero before
            converting those.

        Example
        -------
        For an SEIR model with infection rate ``eta``.

        .. code:: python

            epi.set_quadratic_events([
                ( ("S", "I"),
                  eta,
                  [ ("S", -1), ("E", +1) ]
                ),
            ])

        Read  as

        "Coupling of *S* and *I* leads to
        the decay of one *S* particle to one *E* particle with
        rate :math:`\eta`.".
        """

        if reset_events:
            quadratic_event_updates = []
            quadratic_rate_functions = []
            quadratic_events = []
        else:
            quadratic_event_updates = list(self.quadratic_event_updates)
            quadratic_rate_functions = list(self.quadratic_rate_functions)
            quadratic_events = list(self.quadratic_events)

        for coupling_compartments, rate, affected_compartments in event_list:

            _s0 = self.get_compartment_id(coupling_compartments[0])
            _s1 = self.get_compartment_id(coupling_compartments[1])

            dy = np.zeros(self.N_comp)
            for trg, change in affected_compartments:
                _t = self.get_compartment_id(trg)
                dy[_t] += change

            if self._rate_has_functional_dependency(rate):
                this_rate = DynamicQuadraticRate(rate, _s0, _s1)
            else:
                this_rate = ConstantQuadraticRate(rate, _s0, _s1)

            quadratic_event_updates.append( dy )
            quadratic_rate_functions.append( this_rate )
            quadratic_events.append( (coupling_compartments, rate, affected_compartments) )

        if not allow_nonzero_column_sums and len(quadratic_rate_functions)>0:
            _y = np.ones(self.N_comp)
            test = sum([r(0,_y) * dy for dy, r in zip (quadratic_event_updates, quadratic_rate_functions)])
            test_sum = test.sum()
            if np.abs(test_sum) > 1e-15:
                warnings.warn("events do not sum to zero for each column:" + str(test_sum))

        self.quadratic_event_updates = quadratic_event_updates
        self.quadratic_rate_functions = quadratic_rate_functions
        self.quadratic_events = quadratic_events

        return self

    def dydt(self,t,y):
        """
        Compute the current momenta of the epidemiological model.

        Parameters
        ----------
        t : :obj:`float`
            Current time
        y : numpy.ndarray
            The entries correspond to the compartment frequencies
            (or counts, depending on population size).
        """
        
        ynew = sum([r(t,y) * dy for dy, r in zip(self.linear_event_updates, self.linear_rate_functions)])
        ynew += sum([r(t,y) * dy for dy, r in zip(self.birth_event_updates, self.birth_rate_functions)])
        if self.correct_for_dynamical_population_size:
            population_size = y.sum()
        else:
            population_size = self.initial_population_size
        ynew += sum([r(t,y)/population_size * dy for dy, r in zip(self.quadratic_event_updates, self.quadratic_rate_functions)])

        return ynew

    def get_numerical_dydt(self):
        """
        Return a function that obtains ``t`` and ``y`` as an input and returns ``dydt`` of this system
        """
        return self.dydt

    def get_time_leap_and_proposed_compartment_changes(self,
                                                       t,
                                                       current_event_rates = None, 
                                                       get_event_rates = None,
                                                       get_compartment_changes = None,
                                                       ):
        """
        For the current event rates, obtain a proposed
        time leap and concurrent state change vector.

        This method is needed for stochastic simulations.

        Parameters
        ----------
        t : float
            current time
        current_event_rates : list, default = None
            A list of constant rate values.
            Will be ignored if
            ``self.rates_have_explicit_time_dependence`` is ``True``,
            which is why ``None`` is a valid value.
        get_event_rates : function, default = None
            A function that takes time ``t`` and current
            state ``y`` as input and computes the rates of 
            all possible events.
            If ``None``, will attempt
            to set this to self.get_event_rates().
        get_compartment_changes : function, default = None
            A function that takes computed event rates
            and returns a random state change with
            probability proportional to its rate.
            If ``None``, will attempt
            to set this to self.get_compartment_changes().

        Returns
        -------
        tau : float
            A time leap.
        dy : numpy.ndarray
            A state change vector.
        """

        if get_event_rates is None:
            get_event_rates = self.get_event_rates
        if get_compartment_changes is None:
            get_compartment_changes = self.get_compartment_changes

        if self.rates_have_explicit_time_dependence:
            # solve the integral numerically
            if self.use_ivp_solver:
                new_t = time_leap_ivp(t, self.y0, get_event_rates)
            else:
                new_t = time_leap_newton(t, self.y0, get_event_rates)

            tau = new_t - t
            proposed_event_rates = get_event_rates(new_t, self.y0)
            dy = get_compartment_changes(proposed_event_rates)
        else:

            total_event_rate = current_event_rates.sum()
        
            tau = np.random.exponential(1/total_event_rate)
            dy = get_compartment_changes(current_event_rates)

        return tau, dy

    def get_compartment_changes(self, rates):
        """
        Sample a state change vector with probability
        proportional to its rate in ``rates``.
        
        Needed for stochastic simulations.

        Parameters
        ==========
        rates : numpy.ndarray
            A non-zero list of rates.
            Expects ``rates`` to be sorted according
            to 
            ``self.birth_event_updates + self.linear_event_updates + self.quadratic_event_updates``.

        Returns
        =======
        dy : numpy.ndarray
            A state change vector.
        """

        idy = custom_choice(rates/rates.sum())

        if idy < len(self.birth_event_updates):
            return self.birth_event_updates[idy]
        elif idy < len(self.birth_event_updates) + len(self.linear_event_updates):
            idy -= len(self.birth_event_updates)
            return self.linear_event_updates[idy]
        else:
            idy -= (len(self.birth_event_updates) + len(self.linear_event_updates))
            return self.quadratic_event_updates[idy]


    def get_event_rates(self, t, y):
        """
        Get a list of rate values corresponding to the previously
        set events.

        Parameters
        ----------
        t : float
            Current time
        y : numpy.ndarray
            Current state vector

        Returns
        -------
        rates : list
            A list of rate values corresponding to rates.
            Ordered as ``birth_rate_functions +
            linear_rate_functions + quadratic_rate_functions``.
        """
        rates = [r(t,y) for r in self.birth_rate_functions]
        rates += [r(t,y) for r in self.linear_rate_functions]
        if self.correct_for_dynamical_population_size:
            population_size = self.y0.sum()
        else:
            population_size = self.initial_population_size
        rates += [ r(t,self.y0)/population_size for r in self.quadratic_rate_functions ]
        rates = np.array(rates)

        return rates

    def get_numerical_event_and_rate_functions(self):
        """
        This function is needed to generalize
        stochastic simulations for child classes.

        Returns
        -------
        get_event_rates : func
            A function that takes the current time ``t`` and
            state vector ``y``
            and returns numerical event rate lists.
        get_compartment_changes : funx
            A function that takes a numerical list of event ``rates``
            and returns a random event state change vector
            with probability proportional to its entry in ``rates``.
        """
        return self.get_event_rates, self.get_compartment_changes

    def simulate(self,
                 tmax,
                 return_compartments=None,
                 sampling_dt=None,
                 sampling_callback=None,
                 adopt_final_state=False,
                 ):
        """
        Returns values of the given compartments at the demanded
        time points (as a numpy.ndarray of shape 
        ``(return_compartments), len(time_points)``.

        If ``return_compartments`` is None, all compartments will
        be returned.

        Parameters
        ----------
        tmax : float
            maximum length of the simulation
        return_compartments : list of compartments, default = None:
            The compartments for which to return time series.
            If ``None``, all compartments will be returned.
        sampling_dt : float, default = None
            Temporal distance between samples of the compartment counts.
            If ``None``, every change will be returned.
        sampling_callback : funtion, default = None
            A function that's called when a sample is taken

        Returns
        -------
        t : numpy.ndarray
            times at which compartment counts have been sampled
        result : dict
            Dictionary mapping a compartment to a time series of its count.
        """

        if return_compartments is None:
            return_compartments = self.compartments

        if sampling_callback is not None and sampling_dt is None:
            raise ValueError('A sampling callback function can only be set if sampling_dt is set, as well.')

        ndx = [self.get_compartment_id(C) for C in return_compartments]
        current_state = self.y0.copy()
        compartments = [ current_state.copy() ]

        if not adopt_final_state:
            initial_state = current_state.copy()
            initial_time = self.t0

        t = self.t0
        time = [self.t0]

        get_event_rates, get_compartment_changes = self.get_numerical_event_and_rate_functions()
        current_event_rates = get_event_rates(t, self.y0)
        total_event_rate = current_event_rates.sum()

        if sampling_callback is not None:
            sampling_callback()

        # Check for a) zero event rate and b) zero possibility for any nodes being changed still.
        # This is important because it might happen that nodes
        # have a non-zero reaction rate but no targets left
        # at which point the simulation will never halt.
        while t < tmax and \
              total_event_rate > 0:

            # sample and advance time according to current total rate
            tau, dy = self.get_time_leap_and_proposed_compartment_changes(t,
                                                                          current_event_rates=current_event_rates,
                                                                          get_event_rates=get_event_rates,
                                                                          get_compartment_changes=get_compartment_changes,
                                                                          )
            new_t = t + tau

            # break if simulation time is reached
            if new_t >= tmax:
                break

            # sampling
            if sampling_dt is not None:
                # sample all the time steps that were demanded in between the two events
                last_sample_dt = time[-1]
                for idt in range(1,int(np.ceil((new_t-last_sample_dt)/sampling_dt))):
                    time.append(last_sample_dt+idt*sampling_dt)
                    compartments.append(current_state.copy())
                    if sampling_callback is not None:
                        sampling_callback()

            # write losses and gains into the current state vector
            current_state += dy

            # save the current state if sampling_dt wasn't specified
            if sampling_dt is None:
                time.append(new_t)
                compartments.append(current_state.copy())

            # save current state
            self.t0 = new_t
            self.y0 = current_state.copy()

            current_event_rates = get_event_rates(new_t, self.y0)
            total_event_rate = current_event_rates.sum()

            # advance time
            t = new_t


        if sampling_dt is not None:
            next_sample = time[-1] + sampling_dt
            if next_sample <= tmax:
                time.append(next_sample)
                compartments.append(current_state)
                if sampling_callback is not None:
                    sampling_callback()

        # convert to result dictionary
        time = np.array(time)
        result = np.array(compartments)

        if not adopt_final_state:
            self.y0 = initial_state
            self.t0 = initial_time
        else:
            self.t0 = tmax


        return time, { compartment: result[:,c_ndx] for c_ndx, compartment in zip(ndx, return_compartments) }

class SIModel(EpiModel):
    """
    An SI model derived from :class:`epipack.numeric_epi_models.EpiModel`.
    """

    def __init__(self, infection_rate, initial_population_size=1.0):

        EpiModel.__init__(self, list("SI"), initial_population_size)

        self.set_processes([
                ("S", "I", infection_rate, "I", "I"),
            ])


class SISModel(EpiModel):
    """
    An SIS model derived from :class:`epipack.numeric_epi_models.EpiModel`.

    Parameters
    ----------
    R0 : float
        The basic reproduction number. From this, the infection
        rate is computed as ``infection_rate = R0 * recovery_rate``
    recovery_rate : float
        Recovery rate
    population_size : float, default = 1.0
        Number of people in the population.
    """

    def __init__(self, infection_rate, recovery_rate, initial_population_size=1.0):

        EpiModel.__init__(self, list("SI"), initial_population_size)

        self.set_processes([
                ("S", "I", infection_rate, "I", "I"),
                ("I", recovery_rate, "S" ),
            ])
class SIRModel(EpiModel):
    """
    An SIR model derived from :class:`epipack.numeric_epi_models.EpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, initial_population_size=1.0):

        EpiModel.__init__(self, list("SIR"), initial_population_size)

        self.set_processes([
                ("S", "I", infection_rate, "I", "I"),
                ("I", recovery_rate, "R"),
            ])

class SIRSModel(EpiModel):
    """
    An SIRS model derived from :class:`epipack.numeric_epi_models.EpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, waning_immunity_rate, initial_population_size=1.0):

        EpiModel.__init__(self, list("SIR"), initial_population_size)

        self.set_processes([
                ("S", "I", infection_rate, "I", "I"),
                ("I", recovery_rate, "R"),
                ("R", waning_immunity_rate, "S"),
            ])

class SEIRModel(EpiModel):
    """
    An SEIR model derived from :class:`epipack.numeric_epi_models.EpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, symptomatic_rate, initial_population_size=1.0):

        EpiModel.__init__(self, list("SEIR"), initial_population_size)

        self.set_processes([
                ("S", "I", infection_rate, "E", "I"),
                ("E", symptomatic_rate, "I"),
                ("I", recovery_rate, "R"),
            ])

if __name__=="__main__":    # pragma: no cover
    N = 100
    epi = EpiModel(list("SEIR"),100)
    #print(epi.compartments)
    #print()
    epi.set_processes([
            ("S", "I", 2.0, "E", "I"),
            ("E", 1.0, "I"),
            ("I", 1.0, "R"),
            ])

    print("#printing updates")
    #print([dy.toarray() for dy in epi.linear_event_updates])
    #print([dy.toarray() for dy in epi.quadratic_event_updates])

    import matplotlib.pyplot as pl
    from time import time

    N_meas = 5

    tt = np.linspace(0,20,100)
    start = time()
    epi = EpiModel(list("SEIR"),100)
    epi.set_processes([
                ("S", "I", 2.0, "E", "I"),
                ("E", 1.0, "I"),
                ("I", 1.0, "R"),
            ])
    for meas in range(N_meas):
        epi.set_initial_conditions({'S':0.99*N,'I':0.01*N})
        result = epi.integrate(tt)
    end = time()

    print("arrays needed", end-start,"s")

    pl.plot(tt, result['S'],label='S')
    pl.plot(tt, result['E'],label='E')
    pl.plot(tt, result['I'],label='I')
    pl.plot(tt, result['R'],label='R')
    pl.legend()

    from epipack import MatrixSEIRModel

    tt = np.linspace(0,20,50)
    SEIR = MatrixSEIRModel(2.0,1.0,1.0,initial_population_size=N)
    SEIR.set_initial_conditions({'S':0.99*N,'I':0.01*N})
    result = SEIR.integrate(tt)
    pl.plot(tt, result['S'],'s',label='S',mfc='None')
    pl.plot(tt, result['E'],'s',label='E',mfc='None')
    pl.plot(tt, result['I'],'s',label='I',mfc='None')
    pl.plot(tt, result['R'],'s',label='R',mfc='None')



    ##########
    epi = EpiModel(list("SEIR"),100)
    epi.set_processes([
                ("S", "I", 2.0, "E", "I"),
                ("E", 1.0, "I"),
                ("I", 1.0, "R"),
            ])
    epi.set_initial_conditions({'S':99,'I':1})
    t, result = epi.simulate(tt[-1])
    pl.plot(t, result['S'],label='S')
    pl.plot(t, result['E'],label='E')
    pl.plot(t, result['I'],label='I')
    pl.plot(t, result['R'],label='R')

    pl.figure()

    S, I, R = list("SIR")
    N = 200
    model = EpiModel([S,I],N,correct_for_dynamical_population_size=True)

    def temporalR0(t,y):
        return 4 + np.cos(t/100*2*np.pi)

    model.set_processes([
            (S, I, temporalR0, I, I),
            (None, N, S),
            (I, 1, S),
            (S, 1, None),
            (I, 1, None),
        ])
    model.set_initial_conditions({
            S: 190,
            I: 10,
        })

    def print_status():
        print(model.t0/150*100)
    t, result = model.simulate(150,sampling_dt=0.5,sampling_callback=print_status)
    pl.plot(t, result['S'],label='S')
    pl.plot(t, result['I'],label='I')

    model.set_initial_conditions({
            S: 190,
            I: 10,
        })
    tt = np.linspace(0,100,1000)
    result = model.integrate(tt)
    pl.plot(tt, result['S'],label='S')
    pl.plot(tt, result['I'],label='I')

    pl.show()




import unittest

import numpy as np
from scipy.optimize import root
from scipy.integrate import cumtrapz
from scipy.stats import entropy, poisson

from epipack import StochasticEpiModel
from epipack.temporal_networks import TemporalNetwork, TemporalNetworkSimulation

class FakeTacomaNetwork():

    t = [0,0.5,0.6]
    edges = [ [ (0,1) ], [ (0,1), (0,2) ], [] ]
    N = 3
    tmax = 1.0

class TemporalNetworkTest(unittest.TestCase):

    def test_temporal_network(self):

        expected = [
            (0, 0.5, [(0, 1, 1.0)]),
            (0.5, 0.6, [(0, 1, 1.0), (0, 2, 1.0)]),
            (0.6, 1.0, []),
            (1.0, 1.5, [(0, 1, 1.0)]),
            (1.5, 1.6, [(0, 1, 1.0), (0, 2, 1.0)]),
            (1.6, 2.0, []),
            (2.0, 2.5, [(0, 1, 1.0)]),
            (2.5, 2.6, [(0, 1, 1.0), (0, 2, 1.0)]),
            (2.6, 3.0, []),
        ]

        edges = [ [ (0,1) ], [ (0,1), (0,2) ], [] ]
        temporal_network = TemporalNetwork(3,edges,[0,0.5,0.6],1.0)
        for (edge_list, t, next_t), (_t, _next_t, _edge_list) in zip(temporal_network, expected):
            if t >= 3.0:
                break
            assert(t == _t)
            assert(next_t == _next_t)
            assert(set(edge_list) == set(_edge_list))


    def test_temporal_gillespie(self,plot=False):

        infection_rate = 1.0
        recovery_rate = 0.2
        model = StochasticEpiModel(["S","I","R"],3)\
                    .set_link_transmission_processes([
                            ("I", "S", infection_rate, "I", "I"),
                        ])\
                    .set_node_transition_processes([
                            ("I", recovery_rate, "R"),
                        ])\
                    .set_node_statuses([1,0,0])

        edges = [ [ (0,1) ], [ (0,1), (0,2) ], [] ]
        temporal_network = TemporalNetwork(3,edges,[0,0.5,1.2],1.5)
        sim = TemporalNetworkSimulation(temporal_network, model)
        N_meas = 10000
        taus = []
        for meas in range(N_meas):
            sim.reset()
            t, res = sim.simulate(1000)
            if t[-1] == 0:
                continue
            else:
                taus.append(t[1])


        def rate(t):
            t = t % 1.5
            if t < 0.5:
                return infection_rate + recovery_rate
            elif t < 1.2:
                return 2*infection_rate + recovery_rate
            elif t < 1.5:
                return recovery_rate


        measured, bins = np.histogram(taus,bins=100,density=True)
        rates = np.array([rate(_t) for _t in bins])
        I2 = cumtrapz(rates,bins,initial=0.0)
        theory = [ np.exp(-I2[i-1])-np.exp(-I2[i]) for i in range(1,len(bins)) if measured[i-1] > 0]
        experi = [ measured[i-1] for i in range(1,len(bins)) if measured[i-1] > 0]
        # make sure the kullback-leibler divergence is below some threshold
        if plot: # pragma: no cover
            import matplotlib.pyplot as pl
            pl.figure()
            pl.hist(taus,bins=100,density=True)
            tt = np.linspace(0,max(taus),10000)
            rates = np.array([rate(_t) for _t in tt])
            I2 = cumtrapz(rates,tt,initial=0.0)
            pl.plot(tt, rates*np.exp(-I2))
            pl.yscale('log')
            pl.figure()
            pl.hist(taus,bins=100,density=True)
            pl.plot(tt, rates*np.exp(-I2))
            pl.show()
        assert(entropy(theory, experi) < 0.02)

    def test_degree(self):

        edges = [ [ (0,1) ], [ (0,1), (0,2) ], [] ]
        temporal_network = TemporalNetwork(3,edges,[0,0.5,1.5],3.0)
        k = temporal_network.mean_out_degree()
        expected = 0.5*np.mean([1,1,0])+1.0*np.mean([2,1,1])
        expected /= 3.0
        assert( np.isclose(k, expected))
        temporal_network = TemporalNetwork(3,edges,[0,0.5,1.5],3.0,directed=True,weighted=False,loop_network=False)
        k = temporal_network.mean_out_degree()
        expected = 0.5*np.mean([1,0,0])+1.0*np.mean([2,0,0])
        expected /= 3.0
        assert( np.isclose(k, expected))
        edges = [ [ (0,1,1.0) ], [ (0,1,0.5), (0,2,2.0) ], [] ]
        temporal_network = TemporalNetwork(3,edges,[0,0.5,1.5],3.0,directed=False,weighted=True)
        k = temporal_network.mean_out_degree()
        expected = 0.5*np.mean([1,1,0.])+1.0*np.mean([2.5,0.5,2.0])
        expected /= 3.0
        assert( np.isclose(k, expected))
        temporal_network = TemporalNetwork(3,edges,[0,0.5,1.5],3.0,directed=True,weighted=True)
        k = temporal_network.mean_out_degree()
        expected = 0.5*np.mean([1,0,0.])+1.0*np.mean([2.5,0,0])
        expected /= 3.0
        assert( np.isclose(k, expected))

    def test_tacoma_network(self):

        expected = [
            (0, 0.5, [(0, 1, 1.0)]),
            (0.5, 0.6, [(0, 1, 1.0), (0, 2, 1.0)]),
            (0.6, 1.0, []),
            (1.0, 1.5, [(0, 1, 1.0)]),
            (1.5, 1.6, [(0, 1, 1.0), (0, 2, 1.0)]),
            (1.6, 2.0, []),
            (2.0, 2.5, [(0, 1, 1.0)]),
            (2.5, 2.6, [(0, 1, 1.0), (0, 2, 1.0)]),
            (2.6, 3.0, []),
        ]

        temporal_network = TemporalNetwork.from_tacoma(FakeTacomaNetwork())
        for (edge_list, t, next_t), (_t, _next_t, _edge_list) in zip(temporal_network, expected):
            if t >= 3.0:
                break
            assert(t == _t)
            assert(next_t == _next_t)
            assert(set(edge_list) == set(_edge_list))



if __name__ == "__main__":

    import sys

    T = TemporalNetworkTest()
    T.test_tacoma_network()
    T.test_temporal_network()
    T.test_degree()
    T.test_temporal_gillespie(plot=True)



import unittest

import numpy as np
from scipy.optimize import root
from scipy.stats import entropy, poisson

from epipack.numeric_epi_models import (
            DynamicBirthRate,
            ConstantBirthRate,
            DynamicLinearRate,
            ConstantLinearRate,
            DynamicQuadraticRate,
            ConstantQuadraticRate,
            EpiModel,
            SISModel,
            SIModel,
            SIRModel,
            SEIRModel,
            SIRSModel,
        )

from epipack.integrators import time_leap_ivp, time_leap_newton

from epipack.stochastic_epi_models import StochasticEpiModel

class EpiTest(unittest.TestCase):

    def test_compartments(self):
        epi = EpiModel(list("SEIR"))
        assert(all([ i == epi.get_compartment_id(C) for i, C in enumerate("SEIR") ]))
        assert(epi.get_compartment_id("E") == 1)
        assert(epi.get_compartment(1) == "E")

    def test_linear_rates(self):

        epi = EpiModel(list("SEIR"))
        epi.add_transition_processes([
                ("E", 1.0, "I"),
                ("I", 1.0, "R"),
            ])
        linear_rates = [ ConstantLinearRate(1.0,1), ConstantLinearRate(1.0,2) ] 
        linear_events = [ np.array([0,-1,+1,0]), np.array([0,0,-1,+1.]) ]
        for r0, r1 in zip(linear_rates, epi.linear_rate_functions):
            assert(r0(0,[0.1,0.2,0.3,0.4,0.5]) == r1(0, [0.1,0.2,0.3,0.4,0.5]))
        for e0, e1 in zip(linear_events, epi.linear_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))

        epi = EpiModel(list("SEIR"))
        _r0 = lambda t, y: 2+np.cos(t)
        _r1 = lambda t, y: 2+np.sin(t)
        epi.add_transition_processes([
                ("E", _r0,  "I"),
                ("I", _r1,  "R"),
            ])
        linear_rates = [ DynamicLinearRate(_r0,1), DynamicLinearRate(_r1,2) ] 
        linear_events = [ np.array([0,-1,+1,0]), np.array([0,0,-1,+1.]) ]

        for r0, r1 in zip(linear_rates, epi.linear_rate_functions):
            assert(r0(0,[0.1,0.2,0.3,0.4,0.5]) == r1(0, [0.1,0.2,0.3,0.4,0.5]))
        for e0, e1 in zip(linear_events, epi.linear_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))


    def test_adding_linear_rates(self):

        epi = EpiModel(list("SEIR"))
        epi.set_processes([
                ("E", 1.0, "I"),
                ])

        epi.add_transition_processes([
                ("I", 1.0, "R"),
                ])

        linear_rates = [ ConstantLinearRate(1.0,1), ConstantLinearRate(1.0,2) ] 
        linear_events = [ np.array([0,-1,+1,0]), np.array([0,0,-1,+1.]) ]
        for r0, r1 in zip(linear_rates, epi.linear_rate_functions):
            assert(r0(0,[0.1,0.2,0.3,0.4,0.5]) == r1(0, [0.1,0.2,0.3,0.4,0.5]))
        for e0, e1 in zip(linear_events, epi.linear_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))

    def test_quadratic_processes(self):

        epi = EpiModel(list("SEIAR"))
        quadratic_rates = [ ConstantQuadraticRate(1.0,2,0)] 
        quadratic_events = [ np.array([-1,+1,0,0,0.])]
        epi.add_transmission_processes([
                ("S", "I",  1.0, "I", "E"),
            ])
        for r0, r1 in zip(quadratic_rates, epi.quadratic_rate_functions):
            assert(r0(0,[0.1,0.2,0.3,0.4,0.5]) == r1(0, [0.1,0.2,0.3,0.4,0.5]))
        for e0, e1 in zip(quadratic_events, epi.quadratic_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))


    def test_adding_quadratic_processes(self):

        epi = EpiModel(list("SEIAR"))
        quadratic_rates = [ ConstantQuadraticRate(1.0,2,0), ConstantQuadraticRate(1.0,3,0) ] 
        quadratic_events = [ np.array([-1,+1,0,0,0.]), np.array([-1,+1,0,0,0.]) ]
        epi.set_processes([
                ("S", "I",  1.0, "I", "E"),
            ])
        epi.add_transmission_processes([
                ("S", "A",  1.0, "A", "E"),
            ])
        for r0, r1 in zip(quadratic_rates, epi.quadratic_rate_functions):
            assert(r0(0,[0.1,0.2,0.3,0.4,0.5]) == r1(0, [0.1,0.2,0.3,0.4,0.5]))
        for e0, e1 in zip(quadratic_events, epi.quadratic_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))


    def test_SIS_with_simulation_restart_and_euler(self):
        N = 100
        epi = SISModel(infection_rate=2,recovery_rate=1,initial_population_size=N)
        epi.set_initial_conditions({'S':0.99*N,'I':0.01*N})
        tt = np.linspace(0,100,2)
        result = epi.integrate(tt,['S'])
        assert(np.isclose(result['S'][-1],N/2))

        tt = np.linspace(0,100,1000)
        result = epi.integrate_and_return_by_index(tt,['S'],integrator='euler')
        assert(np.isclose(result[0,-1],N/2))

    def test_repeated_simulation(self):

        N = 100
        epi = SISModel(infection_rate=2,recovery_rate=1,initial_population_size=N)
        epi.set_initial_conditions({'S':0.99*N,'I':0.01*N})
        tt = np.linspace(0,100,100)
        old_t = tt[0]

        for it, t in enumerate(tt[1:]):
            result = epi.integrate_and_return_by_index([old_t,t],integrator='euler',adopt_final_state=True)
            old_t = t

        assert(np.isclose(result[0,-1],N/2))

    def test_birth_death(self):

        epi = EpiModel(list("SIR"))

        R0 = 2
        rho = 1
        mu = 0.2
        eta = R0 * rho

        with self.assertWarns(UserWarning):
            epi.set_processes([
                    ("S", "I", eta, "I", "I"),
                    ("I", rho, "R"),  
                    (None, mu, "S"),
                    ("S", mu, None),
                    ("R", mu, None),
                    ("I", mu, None),
                ])
        epi.set_initial_conditions({'S': 0.8, 'I':0.2 })

        t = [0,1000]
        res = epi.integrate(t)
        assert(np.isclose(res['S'][-1],(mu+rho)/eta))
        assert(np.isclose(res['I'][-1],mu/eta*(eta-mu-rho)/(mu+rho)))

    def test_dynamic_birth(self):

        A = "A"
        epi = EpiModel([A])
        epi.set_initial_conditions({A:1})
        with self.assertWarns(UserWarning):
            epi.set_processes([
                    (None, lambda t, y: 2*t, A),
                ])
        res = epi.integrate([0,5])
        assert(np.isclose(res[A][-1],5**2+1))

    def test_correcting_for_declining_pop_size(self):

        A, B = list("AB")
        epi = EpiModel([A, B],10,correct_for_dynamical_population_size=True)
        epi.add_transition_processes([
                #(None, 0.1, A),
            ])
        epi.add_fusion_processes([
                (A, B, 1, B),
            ])
        epi.set_initial_conditions({B:4, A:6})
        tt = np.linspace(0,30)
        result = epi.integrate(tt)
        #from matplotlib import pyplot as pl
        #pl.plot(tt, result[A], label=A)
        #pl.plot(tt, result[B], label=B)
        epi.correct_for_dynamical_population_size = False
        result = epi.integrate(tt)
        #pl.plot(tt, result[A], label=A)
        #pl.plot(tt, result[B], label=B)
        #pl.legend()
        #pl.show()


    def test_fusion_and_adding_rates(self):

        A, B, C = list("ABC")

        epi = EpiModel(list("ABC"))

        # this should not raise a warning that rates do not sum to zero
        # as it will be actively suppressed
        epi.add_fusion_processes([
                (A, B, 1, C),
            ])

        with self.assertWarns(UserWarning):
            # this should raise a warning that rates do not sum to zero
            epi.add_quadratic_events([
                    ((A, B), 1, [(C, -1),(A, +1)]),
                ])
        # now rates should sum to zero
        epi.add_quadratic_events([
                        ((A, B), 1, [(B, +1)]),
                    ])

        with self.assertWarns(UserWarning):
            # this should raise a warning that rates do not sum to zero
            epi.add_linear_events([
                   ((A,), 1, [(B,-1)]) 
                ])

    def test_initial_condition_warnings(self):

        A, B, C = list("ABC")

        epi = EpiModel(list("ABC"))

        with self.assertWarns(UserWarning):
            # this should raise a warning that rates do not sum to zero
            epi.set_initial_conditions({A:0.1,B:0.2})

        with self.assertWarns(UserWarning):
            # this should raise a warning that initial conditions were set twice
            epi.set_initial_conditions([(A,0.1),(A,0.2)])


    def test_custom_models(self):

        S, I, R = list("SIR")

        eta = 1
        epi = SIModel(eta)
        epi.set_initial_conditions({"S":0.99, "I":0.01})
        epi.integrate([0,1000],adopt_final_state=True)
        assert(np.isclose(epi.y0[0],0))


        
        eta = 2
        rho = 1
        epi = SIRModel(eta,rho)
        S0 = 0.99
        epi.set_initial_conditions({S:S0, I:1-S0})
        R0 = eta/rho
        Rinf = lambda x: 1-x-S0*np.exp(-x*R0)
        res = epi.integrate([0,100])

        SIR_theory = root(Rinf,0.5).x[0]
        assert(np.isclose(res[R][-1],SIR_theory))


        omega = 1
        epi = SEIRModel(eta,rho,omega)
        epi.set_initial_conditions({S:S0, I:1-S0})
        res = epi.integrate([0,100])
        assert(np.isclose(res[R][-1],SIR_theory))
        #======================


        epi = SISModel(eta, rho, initial_population_size=100)

        epi.set_initial_conditions({S: 99, I:1 })

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[S][-1],50))

        epi = SIRSModel(eta, rho, omega)

        epi.set_initial_conditions({S: 0.99, I:0.01 })

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[R][-1],(1-rho/eta)/(1+omega/rho)))

    def test_temporal_gillespie(self,plot=False):

        scl = 40
        def R0(t,y=None):
            return 4+np.cos(t*scl)

        S, I = list("SI")
        N = 100
        rec = 1
        model = EpiModel([S,I], N)
        model.set_processes([
                (S, I, R0, I, I),
                (I, rec, S),
            ])
        I0 = 1
        S0 = N - I0
        model.set_initial_conditions({
                S: S0,
                I: I0,
            })

        taus = []
        N_sample = 10000
        for sample in range(N_sample):
            tau, _ = model.get_time_leap_and_proposed_compartment_changes(0)
            taus.append(tau)

        I = lambda t: (4*t + 1/scl*np.sin(t*scl))
        I2 = lambda t: I(t)*S0*I0/N+I0*rec*t
        pdf = lambda t: (R0(t)*S0*I0/N + I0*rec) * np.exp(-I2(t))
        measured, bins = np.histogram(taus,bins=100,density=True)
        theory = [ np.exp(-I2(bins[i-1]))-np.exp(-I2(bins[i])) for i in range(1,len(bins)) if measured[i-1] > 0]
        experi = [ measured[i-1] for i in range(1,len(bins)) if measured[i-1] > 0]
        # make sure the kullback-leibler divergence is below some threshold
        if plot: # pragma: no cover
            import matplotlib.pyplot as pl
            pl.figure()
            pl.hist(taus,bins=100,density=True)
            tt = np.linspace(0,1,100)
            pl.plot(tt, pdf(tt))
            pl.yscale('log')
            pl.figure()
            pl.hist(taus,bins=100,density=True)
            tt = np.linspace(0,1,100)
            pl.plot(tt, pdf(tt))
            pl.show()
        assert(entropy(theory, experi) < 0.01)

    def test_temporal_gillespie_repeated_simulation(self,plot=False):

        scl = 40
        def R0(t,y=None):
            return 4+np.cos(t*scl)

        S, I = list("SI")
        N = 100
        rec = 1
        model = EpiModel([S,I], N)
        model.set_processes([
                (S, I, R0, I, I),
                (I, rec, S),
            ])
        I0 = 1
        S0 = N - I0
        model.set_initial_conditions({
                S: S0,
                I: I0,
            })

        taus = []
        N_sample = 10000
        if plot:
            from tqdm import tqdm
        else:
            tqdm = lambda x: x
        tt = np.linspace(0,1,100)
        for sample in tqdm(range(N_sample)):
            tau = None
            model.set_initial_conditions({
                    S: S0,
                    I: I0,
                })
            for _t in tt[1:]: 
                time, result = model.simulate(_t,adopt_final_state=True)
                #print(time, result['I'])
                if result['I'][-1] != I0:
                    tau = time[1]
                    break
            #print()
            if tau is not None:
                taus.append(tau)

        I = lambda t: (4*t + 1/scl*np.sin(t*scl))
        I2 = lambda t: I(t)*S0*I0/N+I0*rec*t
        pdf = lambda t: (R0(t)*S0*I0/N + I0*rec) * np.exp(-I2(t))
        measured, bins = np.histogram(taus,bins=100,density=True)
        theory = [ np.exp(-I2(bins[i-1]))-np.exp(-I2(bins[i])) for i in range(1,len(bins)) if measured[i-1] > 0]
        experi = [ measured[i-1] for i in range(1,len(bins)) if measured[i-1] > 0]
        # make sure the kullback-leibler divergence is below some threshold
        if plot:
            import matplotlib.pyplot as pl
            pl.figure()
            pl.hist(taus,bins=100,density=True)
            tt = np.linspace(0,1,100)
            pl.plot(tt, pdf(tt))
            pl.yscale('log')
            pl.figure()
            pl.hist(taus,bins=100,density=True)
            tt = np.linspace(0,1,100)
            pl.plot(tt, pdf(tt))
            pl.show()
        assert(entropy(theory, experi) < 0.01)

    def test_stochastic_well_mixed(self):

        S, E, I, R = list("SEIR")

        N = 75000
        tmax = 100
        model = EpiModel([S,E,I,R],N)
        model.set_processes([
                ( S, I, 2, E, I ),
                ( I, 1, R),
                ( E, 1, I),
            ])
        model.set_initial_conditions({S: N-100, I: 100})

        tt = np.linspace(0,tmax,10000)
        result_int = model.integrate(tt)

        t, result_sim = model.simulate(tmax,sampling_dt=1,return_compartments=[S, R])

        model = StochasticEpiModel([S,E,I,R],N)
        model.set_link_transmission_processes([
                ( I, S, 2, I, E ),
            ])
        model.set_node_transition_processes([
                ( I, 1, R),
                ( E, 1, I),
            ])
        model.set_random_initial_conditions({S: N-100, I: 100})

        t, result_sim2 = model.simulate(tmax,sampling_dt=1,return_compartments=[S, R])

        for c, res in result_sim2.items():
            #print(c, np.abs(1-res[-1]/result_int[c][-1]))
            #print(c, np.abs(1-res[-1]/result_sim[c][-1]))
            assert(np.abs(1-res[-1]/result_int[c][-1]) < 0.05)
            assert(np.abs(1-res[-1]/result_sim[c][-1]) < 0.05)


    def test_stochastic_fission(self):

        A, B, C = list("ABC")

        N = 10
        epi = EpiModel([A,B,C],N,correct_for_dynamical_population_size=True)
        epi.add_fusion_processes([
                (A, B, 1.0, C),
            ])
        epi.set_initial_conditions({ A: 5, B: 5})

        t, res = epi.simulate(1e9)

        assert(res[C][-1] == 5)

    def test_birth_stochastics(self):

        A, B, C = list("ABC")

        epi = EpiModel([A,B,C],10,correct_for_dynamical_population_size=True)
        epi.set_initial_conditions({A:5, B:5})

        epi.set_processes([
                (None, 1, A),
                (A, 1, B),
                (B, 1, None),
            ],allow_nonzero_column_sums=True)

        _, res = epi.simulate(200,sampling_dt=0.05)

        vals = np.concatenate([res[A][_>10], res[B][_>10]])
        rv = poisson(vals.mean())
        measured, bins = np.histogram(vals,bins=np.arange(10)-0.5,density=True)
        theory = [ rv.pmf(i) for i in range(0,len(bins)-1) if measured[i] > 0]
        experi = [ measured[i] for i in range(0,len(bins)-1) if measured[i] > 0]
        # make sure the kullback-leibler divergence is below some threshold
        #for a, b in zip(theory, experi):
        #    print(a,b)
        assert(entropy(theory, experi) < 1e-2)
        assert(np.median(res[A]) == 1)

    def test_sampling_callback(self):
        epi = SIModel(infection_rate=5.0,initial_population_size=100)
        epi.set_initial_conditions({"S":90,"I":10})
        self.assertRaises(ValueError,epi.simulate,1,sampling_callback=lambda x: x)

        i = 0
        samples = []
        def sampled():
            samples.append(epi.y0[0])

        t, res = epi.simulate(10,sampling_dt=0.1,sampling_callback=sampled)

        assert(all([a==b for a, b in zip(res['S'], samples)]))


    def test_integral_solvers(self):

        def get_event_rates(t, y):
            return y * (0.05 + 0.03 * np.array([ np.cos(t), np.sin(t), np.cos(t)**2, np.sin(t)**2 ]))

        rand = 0.834053
        t0 = 1.0
        y0 = np.array([0.1,0.2,0.3,0.4])
        t_nwt = time_leap_newton(t0, y0, get_event_rates, rand)
        t_ivp = time_leap_ivp(t0, y0, get_event_rates, rand)
        expected = 30.76
        numeric = np.array([t_nwt, t_ivp])
        assert(np.all( np.abs(numeric-expected)/numeric < 1e-3) )

    def test_integrate_until(self):

        N = 100
        epi = SIModel(infection_rate=5.0,initial_population_size=N)
        epi.set_initial_conditions({"S":90,"I":10})

        thresh = 0.5
        iS = epi.get_compartment_id("S")
        stop_condition = lambda t, y: thresh*N - y[iS]

        t, res = epi.integrate_until(0,stop_condition,return_compartments=['S'])

        assert(np.isclose(thresh*N,res['S'][-1]))




if __name__ == "__main__":

    import sys

    T = EpiTest()
    T.test_integrate_until()
    T.test_integral_solvers()
    T.test_temporal_gillespie_repeated_simulation()
    T.test_sampling_callback()
    T.test_birth_stochastics()
    T.test_stochastic_fission()
    T.test_correcting_for_declining_pop_size()
    T.test_dynamic_birth()
    T.test_stochastic_well_mixed()
    T.test_temporal_gillespie()
    T.test_compartments()
    T.test_linear_rates()
    T.test_adding_linear_rates()
    T.test_quadratic_processes()
    T.test_adding_quadratic_processes()
    T.test_SIS_with_simulation_restart_and_euler()
    T.test_repeated_simulation()
    T.test_custom_models()
    T.test_birth_death()
    T.test_fusion_and_adding_rates()
    T.test_initial_condition_warnings()



import unittest

import numpy as np
import sympy

from epipack.process_conversions import (
            processes_to_rates,
            transition_processes_to_rates,
            fission_processes_to_rates,
            fusion_processes_to_rates,
            transmission_processes_to_rates,
            processes_to_events,
            transition_processes_to_events,
            fission_processes_to_events,
            fusion_processes_to_events,
            transmission_processes_to_events,
        )

class ProcessConversionTest(unittest.TestCase):

    def test_exceptions(self):

        self.assertRaises(TypeError,processes_to_rates,
                    [("S","S","S")], ["S"]
                )
        self.assertRaises(TypeError,processes_to_rates,
                    [("S","S","S","S")], ["S"]
                )
        self.assertRaises(TypeError,processes_to_rates,
                    [("S","S","S","S","S")], ["S"]
                )
        self.assertRaises(TypeError,processes_to_rates,
                    [("S","S","S","S","S","S")], ["S"]
                )
        self.assertRaises(TypeError,processes_to_events,
                    [("S","S","S")], ["S"]
                )
        self.assertRaises(TypeError,processes_to_events,
                    [("S","S","S","S")], ["S"]
                )
        self.assertRaises(TypeError,processes_to_events,
                    [("S","S","S","S","S")], ["S"]
                )
        self.assertRaises(TypeError,processes_to_events,
                    [("S","S","S","S","S","S")], ["S"]
                )

        S, I = sympy.symbols("S I")

        processes_to_rates(
                    [(S,S,I)], [S], ignore_rate_position_checks = True 
                )
        self.assertRaises(TypeError,processes_to_rates,
                    [(S,S,S,S)], [S], ignore_rate_position_checks = True
                )
        self.assertRaises(ValueError,processes_to_rates,
                    [(S,S,S,S,S)], [S], ignore_rate_position_checks = True
                )
        self.assertRaises(ValueError,processes_to_rates,
                    [(S,I,S,I,S)], [S], ignore_rate_position_checks = True
                )
        processes_to_events(
                    [(S,S,S,S,I)], [S], ignore_rate_position_checks = True
                )
        processes_to_events(
                    [(S,S,I)], [S], ignore_rate_position_checks = True 
                )
        self.assertRaises(TypeError,processes_to_events,
                    [(S,S,S,S)], [S], ignore_rate_position_checks = True
                )
        self.assertRaises(ValueError,processes_to_events,
                    [(S,S,S,S,S)], [S], ignore_rate_position_checks = True
                )
        self.assertRaises(ValueError,processes_to_events,
                    [(S,I,S,I,S)], [S], ignore_rate_position_checks = True
                )
        processes_to_events(
                    [(S,S,S,S,I)], [S], ignore_rate_position_checks = True
                )

    def test_transitions(self):

        self.assertRaises(ValueError,transition_processes_to_rates,
                    [("S",1,"S")]
                )
        self.assertRaises(ValueError,transition_processes_to_rates,
                    [(None,1,None)]
                )
        linear_rates = transition_processes_to_rates([
                ("S", 1, "I"),
                ("S", 1, None),
                (None, 1, "I"),
            ])
        expected = [
                    ("S", "S", -1),
                    ("S", "I", +1),
                    ("S", "S", -1),
                    (None, "I", +1),
                ]

        assert(all(a==b for a, b in zip(linear_rates, expected)))

        self.assertRaises(ValueError,transition_processes_to_events,
                    [("S",1,"S")]
                )
        self.assertRaises(ValueError,transition_processes_to_events,
                    [(None,1,None)]
                )
        linear_events = transition_processes_to_events([
                ("S", 2, "I"),
                ("S", 2, None),
                (None, 2, "I"),
            ])
        expected = [
                    (("S",), 2, [("S", -1), ("I", +1)]),
                    (("S",), 2, [("S", -1), ]),
                    ((None,), 2,[( "I", +1)]),
                ]

        assert(str(linear_events) == str(expected))


    def test_fission(self):

        rates = fission_processes_to_rates([
                ( "A", 3.14, "B", "C"),
            ])
        expected = [
                ( "A", "A", -3.14 ),
                ( "A", "B", +3.14 ),
                ( "A", "C", +3.14 ),
                ]
        assert(all(a==b for a, b in zip(rates, expected)))

        events = fission_processes_to_events([
                ( "A", 3.14, "B", "C"),
            ])
        expected = [
                ( ("A",), 3.14, [("A", -1), ("B", +1), ("C", +1)]),
                ]
        assert(str(events) == str(expected))

    def test_fusion(self):

        rates = fusion_processes_to_rates([
                ( "A", "B", 3.14, "C"),
            ])
        expected = [
                ( "A", "B", "C", +3.14 ),
                ( "A", "B", "A", -3.14 ),
                ( "A", "B", "B", -3.14 ),
                ]
        assert(all(a==b for a, b in zip(rates, expected)))

        events = fusion_processes_to_events([
                ( "A", "B", 3.14, "C"),
            ])
        expected = [
                ( ("A", "B"), 3.14, [("C", +1), ("A", -1), ("B", -1)]),
                ]
        assert(str(events)==str(expected))

    def test_transmission(self):
        rates = transmission_processes_to_rates([
                ( "S", "I", 0.5, "I", "I"),    
                ( "S", "I", 0.5, "B", "C"),    
            ])
        expected = [
                ("I","S","S", -0.5),
                ("I","S","I", +0.5),
                ("S","I","I", -0.5),
                ("S","I","C", +0.5),
                ("S","I","S", -0.5),
                ("S","I","B", +0.5),
            ]
        assert(all(a==b for a, b in zip(rates, expected)))

        events = transmission_processes_to_events([
                ( "S", "I", 0.5, "I", "I"),    
                ( "S", "I", 0.5, "B", "C"),    
            ])
        expected = [
                (("I","S"),0.5,[("S", -1),("I",+1)]),
                (("S","I"),0.5,[("I", -1),("C",+1),("S",-1),("B",+1)]),
            ]

        assert(str(events)==str(expected))

    def test_all(self):
        
        rates = [
                ("S", 1, "I"),
                ("S", 1, None),
                (None, 1, "I"),
            ]
        expected = [
                    ("S", "S", -1),
                    ("S", "I", +1),
                    ("S", "S", -1),
                    (None, "I", +1),
                ]
        rates += [
                ( "A", 3.14, "B", "C"),
            ]
        expected += [
                ( "A", "A", -3.14 ),
                ( "A", "B", +3.14 ),
                ( "A", "C", +3.14 ),
            ]
        rates += [
                ( "A", "B", 3.14, "C"),
            ]
        expected += [
                ( "A", "B", "C", +3.14 ),
                ( "A", "B", "A", -3.14 ),
                ( "A", "B", "B", -3.14 ),
            ]

        rates += [
                ( "S", "I", 0.5, "I", "I"),    
                ( "S", "I", 0.5, "B", "C"),    
            ]
        expected += [
                ("I","S","S", -0.5),
                ("I","S","I", +0.5),
                ("S","I","I", -0.5),
                ("S","I","C", +0.5),
                ("S","I","S", -0.5),
                ("S","I","B", +0.5),
            ]

        qrates, lrates = processes_to_rates(rates,["S","I","A","B","C"])
        rates = set(lrates + qrates)
        expected = set(expected)

        assert(rates == expected)

        events = [
                ("S", 2, "I"),
                ("S", 2, None),
                (None, 2, "I"),
            ]
        lexpected = [
                    (("S",), 2, [("S", -1), ("I", +1)]),
                    (("S",), 2, [("S", -1), ]),
                    ((None,), 2,[( "I", +1)]),
                ]
        events += [
                ( "A", 3.14, "B", "C"),
            ]
        lexpected += [
                ( ("A",), 3.14, [("A", -1), ("B", +1), ("C", +1)]),
            ]
        events += [
                ( "A", "B", 3.14, "C"),
            ]
        qexpected = [
                ( ("A", "B"), 3.14, [("C", +1), ("A", -1), ("B", -1)]),
            ]

        events += [
                ( "S", "I", 0.5, "I", "I"),    
                ( "S", "I", 0.5, "B", "C"),    
            ]
        qexpected += [
                (("I","S"),0.5,[("S", -1),("I",+1)]),
                (("S","I"),0.5,[("I", -1),("C",+1),("S",-1),("B",+1)]),
            ]

        qevents, levents = processes_to_events(events,["S","I","A","B","C"])
        events = levents + qevents
        expected = lexpected + qexpected

        assert(str(events) == str(expected))




if __name__ == "__main__":

    T = ProcessConversionTest()
    T.test_exceptions()
    T.test_transitions()
    T.test_fission()
    T.test_fusion()
    T.test_transmission()
    T.test_all()



import unittest

import numpy as np
from scipy.optimize import root

from epipack.numeric_matrix_epi_models import (
            MatrixEpiModel,
            MatrixSISModel,
            MatrixSIModel,
            MatrixSIRModel,
            MatrixSEIRModel,
            MatrixSIRSModel,
        )

class MatrixEpiTest(unittest.TestCase):

    def test_compartments(self):
        epi = MatrixEpiModel(list("SEIR"))
        assert(all([ i == epi.get_compartment_id(C) for i, C in enumerate("SEIR") ]))
        assert(epi.get_compartment_id("E") == 1)
        assert(epi.get_compartment(1) == "E")

    def test_linear_rates(self):

        epi = MatrixEpiModel(list("SEIR"))
        epi.add_transition_processes([
                ("E", 1.0, "I"),
                ("I", 1.0, "R"),
                ])

        linear_rates = np.zeros((4,4))
        linear_rates[1,1] = -1
        linear_rates[2,1] = +1
        linear_rates[2,2] = -1
        linear_rates[3,2] = +1
        assert(np.all(epi.linear_rates.toarray().flatten()==linear_rates.flatten()))

    def test_adding_linear_rates(self):

        epi = MatrixEpiModel(list("SEIR"))
        epi.set_processes([
                ("E", 1.0, "I"),
                ])

        epi.add_transition_processes([
                ("I", 1.0, "R"),
                ])

        linear_rates = np.zeros((4,4))
        linear_rates[1,1] = -1
        linear_rates[2,1] = +1
        linear_rates[2,2] = -1
        linear_rates[3,2] = +1
        assert(np.all(epi.linear_rates.toarray().flatten()==linear_rates.flatten()))

    def test_quadratic_processes(self):

        epi = MatrixEpiModel(list("SEIR"))
        Q = [ np.zeros((4,4)) for C in epi.compartments ]
        Q[0][0,2] = -1
        Q[1][0,2] = +1
        epi.add_transmission_processes([
                ("S", "I",  1.0, "I", "E"),
                ])
        for iM, M in enumerate(epi.quadratic_rates):
            assert(np.all(M.toarray().flatten()==Q[iM].flatten()))

    def test_adding_quadratic_processes(self):

        epi = MatrixEpiModel(list("SEIAR"))
        Q = [ np.zeros((5,5)) for C in epi.compartments ]
        Q[0][0,2] = -1
        Q[0][0,3] = -1
        Q[1][0,2] = +1
        Q[1][0,3] = +1
        epi.set_processes([
                ("S", "I",  1.0, "I", "E"),
                ])
        epi.add_transmission_processes([
                ("S", "A",  1.0, "A", "E"),
                ])
        for iM, M in enumerate(epi.quadratic_rates):
            assert(np.all(M.toarray().flatten()==Q[iM].flatten()))

    def test_SIS_with_simulation_restart_and_euler(self):
        N = 100
        epi = MatrixSISModel(R0=2,recovery_rate=1,initial_population_size=N)
        epi.set_initial_conditions({'S':0.99*N,'I':0.01*N})
        tt = np.linspace(0,100,2)
        result = epi.integrate(tt,['S'])
        assert(np.isclose(result['S'][-1],N/2))

        tt = np.linspace(0,100,1000)
        result = epi.integrate_and_return_by_index(tt,['S'],integrator='euler')
        assert(np.isclose(result[0,-1],N/2))

    def test_repeated_simulation(self):

        N = 100
        epi = MatrixSISModel(R0=2,recovery_rate=1,initial_population_size=N)
        epi.set_initial_conditions({'S':0.99*N,'I':0.01*N})
        tt = np.linspace(0,100,100)
        old_t = tt[0]

        for it, t in enumerate(tt[1:]):
            result = epi.integrate_and_return_by_index([old_t,t],integrator='euler',adopt_final_state=True)
            old_t = t

        assert(np.isclose(result[0,-1],N/2))

    def test_custom_models(self):

        S, I, R = list("SIR")

        eta = 1
        epi = MatrixSIModel(eta)
        epi.set_initial_conditions({"S":0.99, "I":0.01})
        epi.integrate([0,1000],adopt_final_state=True)
        assert(np.isclose(epi.y0[0],0))



        eta = 2
        rho = 1
        epi = MatrixSIRModel(eta,rho)
        S0 = 0.99
        epi.set_initial_conditions({S:S0, I:1-S0})
        R0 = eta/rho
        Rinf = lambda x: 1-x-S0*np.exp(-x*R0)
        res = epi.integrate([0,100])

        SIR_theory = root(Rinf,0.5).x[0]
        assert(np.isclose(res[R][-1],SIR_theory))


        omega = 1
        epi = MatrixSEIRModel(eta,rho,omega)
        epi.set_initial_conditions({S:S0, I:1-S0})
        res = epi.integrate([0,100])
        assert(np.isclose(res[R][-1],SIR_theory))
        #======================


        epi = MatrixSISModel(eta, rho, initial_population_size=100)

        epi.set_initial_conditions({S: 99, I:1 })

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[S][-1],50))

        epi = MatrixSIRSModel(eta, rho, omega)

        epi.set_initial_conditions({S: 0.99, I:0.01 })

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[R][-1],(1-rho/eta)/(1+omega/rho)))

    def test_birth_death(self):

        epi = MatrixEpiModel(list("SIR"))

        R0 = 2
        rho = 1
        mu = 0.2
        eta = R0 * rho

        with self.assertWarns(UserWarning):
            epi.set_processes([
                    ("S", "I", eta, "I", "I"),
                    ("I", rho, "R"),
                    (None, mu, "S"),
                    ("S", mu, None),
                    ("R", mu, None),
                    ("I", mu, None),
                ])
        epi.set_initial_conditions({'S': 0.8, 'I':0.2 })

        t = [0,1000]
        res = epi.integrate(t)
        assert(np.isclose(res['S'][-1],(mu+rho)/eta))
        assert(np.isclose(res['I'][-1],mu/eta*(eta-mu-rho)/(mu+rho)))


    def test_fusion_and_adding_rates(self):

        A, B, C = list("ABC")

        epi = MatrixEpiModel(list("ABC"))

        # this should not raise a warning that rates do not sum to zero
        # as it will be actively suppressed
        epi.add_fusion_processes([
                (A, B, 1, C),
            ])

        with self.assertWarns(UserWarning):
            # this should raise a warning that rates do not sum to zero
            epi.add_quadratic_rates([
                    (A, B, C, -1),
                    (A, B, A, +1),
                ])
        # now rates should sum to zero
        epi.add_quadratic_rates([
                    (A, B, B, +1),
                    ])

        with self.assertWarns(UserWarning):
            # this should raise a warning that rates do not sum to zero
            epi.add_linear_rates([
                   (A, B, -1) 
                ])

    def test_initial_condition_warnings(self):

        A, B, C = list("ABC")

        epi = MatrixEpiModel(list("ABC"))

        with self.assertWarns(UserWarning):
            # this should raise a warning that rates do not sum to zero
            epi.set_initial_conditions({A:0.1,B:0.2})

        with self.assertWarns(UserWarning):
            # this should raise a warning that initial conditions were set twice
            epi.set_initial_conditions([(A,0.1),(A,0.2)])

    def test_R0(self):
        """
        R0 computation test based on

        "The construction of next-generation matrices for compartmental epidemic models"
        by Diekmann, Heesterbeek, Roberts, J. R. Soc. Interface (2010) 7, 873–885
        doi.org/10.1098/rsif.2009.0386

        (Section 2.2)
        """

        b11 = 1
        b12 = 2
        b21 = 3
        b22 = 6

        #we'll use b11*b22 = b12*b21 because it makes 
        #our lives easier in computing R0 analytically

        mu = 4
        g1 = 5
        g2 = 7
        nu1 = 8
        nu2 = 9
        p = 0.2
        N = 100

        base_comps = list("SEIR")
        cats = [1,2]
        comps = []
        for C in base_comps:
            for c in cats:
                comps.append(C+str(c))

        model = MatrixEpiModel(comps,initial_population_size=N)
        transition_processes = [
                    ("S1", mu, None),
                    ("E1", mu, None),
                    ("I1", mu, None),
                    ("R1", mu, None),
                    ("S2", mu, None),
                    ("E2", mu, None),
                    ("I2", mu, None),
                    ("R2", mu, None),
                    ("E1", nu1, "I1"),
                    ("E2", nu2, "I2"),
                    ("I1", g1, "R1"),
                    ("I2", g2, "R2"),
                ]
        transmission_processes = [
                    ( "S1", "I1", b11, "E1", "I1"),
                    ( "S1", "I2", b12, "E1", "I2"),
                    ( "S2", "I1", b21, "E2", "I1"),
                    ( "S2", "I2", b22, "E2", "I2"),
                ]

        birth_processes =[
                (None, p*mu*N, "S1"),
                (None, (1-p)*mu*N, "S2"),
            ]

        model.set_processes(transition_processes + \
                            transmission_processes + \
                            birth_processes,
                            allow_nonzero_column_sums=True,
                            )
        model.set_initial_conditions({
                'S1': p*N,
                'S2': (1-p)*N,
            })

        R0 = model.get_next_generation_matrix_leading_eigenvalue()

        R0theory = p*b11*nu1/(nu1 + mu)/(g1+mu) + (1-p)*b22*nu2/(nu2+mu)/(g2+mu)

        assert(np.isclose(R0,R0theory))

    def test_jacobian(self):
        R0 = 3
        beta = 4
        model = MatrixSIRModel(R0,beta)\
                    .set_initial_conditions({'S': 1,})

        j = model.get_jacobian_leading_eigenvalue()
        modelR0 = model.get_next_generation_matrix_leading_eigenvalue()
        # in SIR model, leading eigenvalue of jacobian must be equal to
        # j = beta*(R-1)
        assert(np.isclose(R0, modelR0))
        assert(np.isclose(np.real(j), beta*(modelR0-1)))

    def test_2times2_NGM(self):
        R0 = 3
        beta = 4
        nu = 2
        model = MatrixSEIRModel(R0,beta,nu)\
                    .set_initial_conditions({'S': 1,})

        modelR0 = model.get_next_generation_matrix_leading_eigenvalue()
        assert(np.isclose(R0, modelR0))



if __name__ == "__main__":

    T = MatrixEpiTest()
    T.test_R0()
    T.test_jacobian()
    T.test_2times2_NGM()
    T.test_compartments()
    T.test_linear_rates()
    T.test_adding_linear_rates()
    T.test_quadratic_processes()
    T.test_adding_quadratic_processes()
    T.test_SIS_with_simulation_restart_and_euler()
    T.test_repeated_simulation()
    T.test_custom_models()
    T.test_birth_death()
    T.test_fusion_and_adding_rates()
    T.test_initial_condition_warnings()



import unittest

import numpy as np

from epipack.mock_samplable_set import MockSamplableSet, choice

class MSSTest(unittest.TestCase):

    def test_all(self):

        s = MockSamplableSet(1.0,2.0,{0:1.2,3:1.8,})

        assert(all(i==j for i, j in zip([0,3],s.items)))
        assert(all(i==j for i, j in zip([1.2,1.8],s.weights)))
        assert(len(s) == 2)
        assert(s.total_weight() == 3.0)
    
        expected = [
            (3, 1.8),
            (3, 1.8),
            (0, 1.2),
            (0, 1.2),
            (0, 1.2),
            ]
        samples = []
        np.random.seed(1)
        for _ in range(5):
            a = s.sample()
            samples.append(a)

        assert(all([
                    i[0] == j[0] and i[1] == j[1] \
                    for i, j in zip(expected, samples)
                ]))

        a = s[0]
        b = s[0]
        assert(a==b)

        s[0] = 2
        assert(all(i==j for i, j in zip([0,3],s.items)))
        assert(all(i==j for i, j in zip([2.,1.8],s.weights)))
        assert(s.total_weight() == 3.8)

        s[1] = 1.3
        assert(all(i==j for i, j in zip([0,1,3],s.items)))
        assert(all(i==j for i, j in zip([2.,1.3,1.8],s.weights)))
        assert(s.total_weight() == 5.1)

        del s[3]
        assert(all(i==j for i, j in zip([0,1],s.items)))
        assert(all(i==j for i, j in zip([2.,1.3],s.weights)))
        assert(s.total_weight() == 3.3)

        exp = [
                (0, 2.0),
                (1, 1.3),
              ]
        samples = [ (item, weight) for item, weight in s ]
        assert(set(exp) == set(samples))

        assert(0 in s)
        assert(45 not in s)

    def test_choice(self):
        N = 5
        a = np.arange(N)
        p = np.arange(N,dtype=float)
        p /= p.sum()
        N_samples = 20000
        samples = np.array([ choice(a,p) for _ in range(N_samples) ])
        count = np.around([ np.count_nonzero(samples==elem)/N_samples for elem in a ],1)
        assert(np.allclose(count,p))
        samples = np.array([ choice(N,p) for _ in range(N_samples) ])
        count = np.around([ np.count_nonzero(samples==elem)/N_samples for elem in a ],1)
        assert(np.allclose(count,p))


    def test_errors(self):

        self.assertRaises(ValueError, MockSamplableSet,
                    1.0,2.0,{0:0.5}
                    )

        self.assertRaises(ValueError, MockSamplableSet,
                    1.0,2.0,{0:2.5}
                    )

        s = MockSamplableSet(1.0,2.0,{0:1.2,3:1.8,})

        self.assertRaises(ValueError, s.__setitem__, 0, 0.5)
        self.assertRaises(ValueError, s.__setitem__, 0, 2.5)
        self.assertRaises(KeyError, s.__getitem__, 2)
        s.clear()




if __name__ == "__main__":

    T = MSSTest()
    T.test_all()
    T.test_choice()
    T.test_errors()



import unittest

import numpy as np
import sympy
from sympy import symbols, Matrix, Eq, sqrt, FiniteSet, Derivative

from scipy.optimize import root
from scipy.stats import poisson, entropy

from epipack import (
            SymbolicEpiModel,
            SymbolicODEModel,
            SymbolicSISModel,
            SymbolicSIRModel,
            SymbolicSIRSModel,
            SymbolicSIModel,
            get_temporal_interpolation,
            MatrixSISModel,
            MatrixEpiModel,
            SymbolicMatrixEpiModel,
        )

class SymbolicEpiTest(unittest.TestCase):

    def test_compartments(self):
        comps = sympy.symbols("S E I R")
        epi = SymbolicEpiModel(comps)
        assert(all([ i == epi.get_compartment_id(C) for i, C in enumerate(comps) ]))

    def test_linear_rates(self):

        S, E, I, R = sympy.symbols("S E I R")
        epi = SymbolicEpiModel([S,E,I,R])
        epi.add_transition_processes([
                (E, 1, I),
                (I, 1, R),
                ])
        linear_rates = [ E, I ] 
        linear_events = [ Matrix([[0,-1,+1,0]]), Matrix([[0,0,-1,+1]]) ]
        for r0, r1 in zip(linear_rates, epi.linear_rate_functions):
            assert(r0 == r1)            
        for e0, e1 in zip(linear_events, epi.linear_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))


    def test_adding_linear_rates(self):

        S, E, I, R = sympy.symbols("S E I R")
        epi = SymbolicEpiModel([S,E,I,R])

        epi.set_processes([
                (E, 1, I),
                ])

        epi.add_transition_processes([
                (I, 1, R),
                ])
        linear_rates = [ E, I ] 
        linear_events = [ Matrix([[0,-1,+1,0]]), Matrix([[0,0,-1,+1]]) ]
        for r0, r1 in zip(linear_rates, epi.linear_rate_functions):
            assert(r0 == r1)            
        for e0, e1 in zip(linear_events, epi.linear_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))

    def test_quadratic_processes(self):

        S, E, I, R = sympy.symbols("S E I R")
        epi = SymbolicEpiModel([S,E,I,R])
        quadratic_rates = [ I*S ] 
        quadratic_events = [ Matrix([[-1,+1,0,0]]) ]
        epi.add_transmission_processes([
                (S, I, 1, I, E),
            ])
        for r0, r1 in zip(quadratic_rates, epi.quadratic_rate_functions):
            assert(r0 == r1)            
        for e0, e1 in zip(quadratic_events, epi.quadratic_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))

    def test_adding_quadratic_processes(self):

        S, E, I, A, R, rho = sympy.symbols("S E I A R rho")
        epi = SymbolicEpiModel([S,E,I,A,R])

        epi.set_processes([
                (S, I, rho, I, E),
                ])
        epi.add_transmission_processes([
                (S, A, rho, A, E),
                ])
        quadratic_rates = [ I*S*rho, S*A*rho ] 
        quadratic_events = [ Matrix([[-1,+1,0,0,0]]), Matrix([[-1,+1,0,0,0]]) ]
        for r0, r1 in zip(quadratic_rates, epi.quadratic_rate_functions):
            assert(r0 == r1)            
        for e0, e1 in zip(quadratic_events, epi.quadratic_event_updates):
            assert(all([_e0==_e1 for _e0, _e1 in zip(e0, e1)]))

    def test_SIS_with_simulation_restart_and_euler(self):
        S, I = sympy.symbols("S I")
        N = 100
        epi = SymbolicSISModel(infection_rate=2,recovery_rate=1,initial_population_size=N)
        epi.set_initial_conditions({S:0.99*N,I:0.01*N})
        tt = np.linspace(0,100,2)
        result = epi.integrate(tt,[S])
        assert(np.isclose(result[S][-1],N/2))

        tt = np.linspace(0,100,1000)
        result = epi.integrate_and_return_by_index(tt,[S],integrator='euler')
        assert(np.isclose(result[0,-1],N/2))


    def test_basic_analytics(self):
        S, I, R, eta, rho, omega, t = symbols("S I R eta rho omega, t")

        SIRS = SymbolicEpiModel([S,I,R])

        SIRS.set_processes([
            #### transmission process ####
            # S + I (eta)-> I + I
            (S, I, eta, I, I),

            #### transition processes ####
            # I (rho)-> R
            # R (omega)-> S
            (I, rho, R),
            (R, omega, S),

        ])

        odes = SIRS.ODEs()

        expected = [Eq(Derivative(S, t), -I*S*eta + R*omega), Eq(Derivative(I, t), I*(S*eta - rho)), Eq(Derivative(R, t), I*rho - R*omega)]

        assert(all([got==exp for got, exp in zip(odes,expected) ]))

        fixed_points = SIRS.find_fixed_points()
        expected = FiniteSet((S, 0, 0), (rho/eta, R*omega/rho, R))

        assert(all([got==exp for got, exp in zip(fixed_points,expected) ]))

        J = SIRS.jacobian()
        expected = Matrix([[-I*eta, -S*eta, omega], [I*eta, S*eta - rho, 0], [0, rho, -omega]])
        N = SIRS.N_comp
        assert(all([J[i,j]==expected[i,j] for i in range(N) for j in range(N)]))

        eig = SIRS.get_eigenvalues_at_disease_free_state()
        expected = {-omega: 1, eta - rho: 1, 0: 1}

        assert(all([v == expected[k] for k, v in eig.items()]))

    def test_numerical(self):

        S, I, eta, rho = sympy.symbols("S I eta rho")

        epi = SymbolicSISModel(eta, rho)

        epi.set_initial_conditions({S: 1-0.01, I:0.01 })
        epi.set_parameter_values({eta:2})

        self.assertRaises(ValueError, epi.integrate,
                    [0,1]
                    )

        epi = SymbolicSISModel(eta, rho)

        epi.set_initial_conditions({S: 1-0.01, I:0.01 })
        epi.set_parameter_values({eta:2,rho:1})

        tt = np.linspace(0,10,1000)
        result = epi.integrate(tt)

        epi2 = MatrixSISModel(2, 1)

        epi2.set_initial_conditions({"S": 1-0.01, "I":0.01 })

        tt = np.linspace(0,10,1000)
        result2 = epi2.integrate(tt)

        for c0, res0 in result.items():
            assert(np.allclose(res0, result2[str(c0)]))


    def test_time_dependent_rates(self):

        B, t = sympy.symbols("B t")
        epi = SymbolicEpiModel([B])
        epi.add_fission_processes([
                (B, t, B, B),
            ])
        epi.set_initial_conditions({B:1})
        result = epi.integrate([0,3],adopt_final_state=True)
        assert(np.isclose(epi.y0[0], np.exp(3**2/2)))
        epi.set_initial_conditions({B:1})
        result = epi.integrate(np.linspace(0,3,10000),integrator='euler',adopt_final_state=True)
        eul = epi.y0[0]
        real = np.exp(3**2/2)
        assert(np.abs(1-eul/real)<1e-2)

    def test_exceptions(self):

        B, mu, t = sympy.symbols("B mu t")
        epi = SymbolicEpiModel([B])
        epi.add_fission_processes([
                (B, mu, B, B),
            ])
        epi.set_initial_conditions({B:1})

        self.assertRaises(ValueError,epi.integrate,[0,1])

        self.assertRaises(ValueError,SymbolicEpiModel,[t])

        self.assertRaises(ValueError,epi.get_eigenvalues_at_disease_free_state)

    def test_custom_models(self):
        eta = sympy.symbols("eta")
        epi = SymbolicSIModel(eta)
        epi.set_parameter_values({eta:1})
        epi.set_initial_conditions({epi.compartments[0]:0.99, epi.compartments[1]:0.01})
        epi.integrate([0,1000],adopt_final_state=True)
        assert(np.isclose(epi.y0[0],0))


        S, I, R, eta, rho = sympy.symbols("S I R eta rho")
        epi = SymbolicSIRModel(eta,rho)
        epi.set_parameter_values({eta:2,rho:1})
        S0 = 0.99
        epi.set_initial_conditions({S:S0, I:1-S0})
        R0 = 2
        Rinf = lambda x: 1-x-S0*np.exp(-x*R0)
        res = epi.integrate([0,1000])

        theory = root(Rinf,0.5)
        assert(np.isclose(res[R][-1],theory.x[0]))

        epi = SymbolicSISModel(eta, rho, initial_population_size=100)

        epi.set_initial_conditions({S: 99, I:1 })
        epi.set_parameter_values({eta:2,rho:1})

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[S][-1],50))

        S, I, R, eta, rho, omega = sympy.symbols("S I R eta rho omega")
        epi = SymbolicSIRSModel(eta, rho, omega)
        _eta = 2
        _rho = 1
        _omega = 1

        epi.set_initial_conditions({S: 0.99, I:0.01 })
        epi.set_parameter_values({eta:_eta, rho:_rho, omega:_omega})

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[R][-1],(1-_rho/_eta)/(1+_omega/_rho)))

    def test_ODEs_jupyter(self):
        
        eta = sympy.symbols("eta")
        epi = SymbolicSIModel(eta)
        epi.ODEs_jupyter()

    def test_temporal_interpolation(self):

        self.assertRaises(ValueError, get_temporal_interpolation, [0,1],[0,1],interpolation_degree=0)

        t = sympy.symbols("t")
        p = sympy.Piecewise((1, (0 <= t) & (t < 1)), (2,(1<=t) & (t<2)))
        p2 = get_temporal_interpolation([0,1,2],[1,2],0)
        assert(p.subs(t,0.5)==p2.subs(t,0.5))
        assert(p.subs(t,1.5)==p2.subs(t,1.5))
        assert(str(p) == str(p2))

        p = sympy.interpolating_spline(1,t,[0,1,2],[1,2,3])
        p2 = get_temporal_interpolation([0,1,2],[1,2,3],1)
        assert(p.subs(t,0.5)==p2.subs(t,0.5))
        assert(p.subs(t,1.5)==p2.subs(t,1.5))
        assert(str(p) == str(p2))

    def test_changing_population_size(self):

        A, B, C, t = sympy.symbols("A B C t")
        epi = SymbolicEpiModel([A,B,C],10,correct_for_dynamical_population_size=True)
        epi.set_initial_conditions({A:5, B:5})
        epi.set_processes([
                (A, B, 1, C),
            ],allow_nonzero_column_sums=True)

        dydt = epi.dydt()
        assert(dydt[0] == -1*A*B/(A + B + C))

        _, res = epi.simulate(1e9)
        assert(res[C][-1] == 5)

        epi.set_processes([
                (None, 1+sympy.log(1+t), A),
                (A, 1+sympy.log(1+t), B),
                (B, 1+sympy.log(1+t), None),
            ],allow_nonzero_column_sums=True)

        rates, comp_changes = epi.get_numerical_event_and_rate_functions()
        _, res = epi.simulate(200,sampling_dt=0.05)

        vals = np.concatenate([res[A][_>10], res[B][_>10]])
        rv = poisson(vals.mean())
        measured, bins = np.histogram(vals,bins=np.arange(10)-0.5,density=True)
        theory = [ rv.pmf(i) for i in range(0,len(bins)-1) if measured[i] > 0]
        experi = [ measured[i] for i in range(0,len(bins)-1) if measured[i] > 0]
        # make sure the kullback-leibler divergence is below some threshold
        assert(entropy(theory, experi) < 2e-3)
        assert(np.median(res[A]) == 1)



    def test_stochastic_well_mixed(self):

        S, E, I, R = sympy.symbols("S E I R")

        N = 75000
        tmax = 100
        model = SymbolicEpiModel([S,E,I,R],N)
        model.set_processes([
                ( S, I, 2, E, I ),
                ( I, 1, R),
                ( E, 1, I),
            ])
        model.set_initial_conditions({S: N-100, I: 100})

        tt = np.linspace(0,tmax,10)
        result_int = model.integrate(tt)

        t, result_sim = model.simulate(tmax,sampling_dt=1,return_compartments=[S, R])

        for c, res in result_sim.items():
            #print(c, np.abs(1-res[-1]/result_int[c][-1]))
            #print(c, np.abs(1-res[-1]/result_sim[c][-1]))
            assert(np.abs(1-res[-1]/result_int[c][-1]) < 0.05)

    def test_ODE_model(self):

        eta, rho = sympy.symbols("eta rho")

        orig_model = SymbolicSISModel(eta, rho)

        ode_model = SymbolicODEModel(orig_model.ODEs())

        assert(all([eq0 == eq1 for eq0, eq1 in zip(orig_model.ODEs(), ode_model.ODEs())]))

        self.assertRaises(AttributeError, ode_model.simulate)
        self.assertRaises(AttributeError, ode_model.set_linear_events)
        self.assertRaises(AttributeError, ode_model.set_quadratic_events)
        self.assertRaises(AttributeError, ode_model.set_processes)
        self.assertRaises(AttributeError, ode_model.add_linear_events)
        self.assertRaises(AttributeError, ode_model.add_quadratic_events)
        self.assertRaises(AttributeError, ode_model.add_transition_processes)
        self.assertRaises(AttributeError, ode_model.add_fission_processes)
        self.assertRaises(AttributeError, ode_model.add_fusion_processes)
        self.assertRaises(AttributeError, ode_model.add_transmission_processes)




if __name__ == "__main__":

    T = SymbolicEpiTest()
    T.test_ODE_model()
    sys.exit(0)
    T.test_changing_population_size()
    T.test_stochastic_well_mixed()
    T.test_SIS_with_simulation_restart_and_euler()
    T.test_compartments()
    T.test_linear_rates()
    T.test_adding_linear_rates()
    T.test_quadratic_processes()
    T.test_adding_quadratic_processes()
    T.test_basic_analytics()
    T.test_numerical()
    T.test_time_dependent_rates()
    T.test_exceptions()
    T.test_custom_models()
    T.test_ODEs_jupyter()
    T.test_temporal_interpolation()



"""
Provides an API to define epidemiological models in terms of sympy symbolic expressions based on a matrix description.
"""

import warnings

import numpy as np 
import scipy.sparse as sprs

import sympy

from epipack.process_conversions import (
            processes_to_rates,
            transition_processes_to_rates,
            fission_processes_to_rates,
            fusion_processes_to_rates,
            transmission_processes_to_rates,
        )

from epipack.numeric_matrix_epi_models import MatrixEpiModel
from epipack.symbolic_epi_models import SymbolicMixin

class SymbolicMatrixEpiModel(SymbolicMixin,MatrixEpiModel):
    """
    A general class to define standard
    mean-field compartmental
    epidemiological model.

    Parameters
    ----------
    compartments : :obj:`list` of :obj:`string`
        A list containing compartment strings.

    Attributes
    ----------
    compartments : :obj:`list` of :obj:`string`
        A list containing strings that describe each compartment,
        (e.g. "S", "I", etc.).
    N_comp : :obj:`int`
        Number of compartments (including population number)
    linear_rates : sympy.Matrix
        Matrix containing
        transition rates of the linear processes.
    quadratic_rates : list of sympy.Matrix
        List of matrices that contain
        transition rates of the quadratic processes
        for each compartment.
    affected_by_quadratic_process : :obj:`list` of :obj:`int`
        List of integer compartment IDs, collecting
        compartments that are affected
        by the quadratic processes

    Example
    -------

    .. code:: python

        >>> epi = SymbolicMatrixEpiModel(symbols("S I R"))
        >>> print(epi.compartments)
        [ S, I, R ]

    """

    def __init__(self,compartments,initial_population_size=1,correct_for_dynamical_population_size=False):
        """
        """
        MatrixEpiModel.__init__(self, compartments, initial_population_size, correct_for_dynamical_population_size)


        self.t = sympy.symbols("t")
        if self.t in self.compartments:
            raise ValueError("Don't use `t` as a compartment symbol, as it is reserved for time.")

        self.has_functional_rates = False

        self.birth_rates = sympy.zeros(self.N_comp,1)
        self.linear_rates = sympy.zeros(self.N_comp, self.N_comp)
        self.quadratic_rates = [ sympy.zeros(self.N_comp, self.N_comp)\
                                 for c in range(self.N_comp) ]
        self.birth_events = sympy.zeros(self.N_comp,1)
        self.linear_events = sympy.zeros(self.N_comp, self.N_comp)
        self.quadratic_events = [ sympy.zeros(self.N_comp, self.N_comp)\
                                 for c in range(self.N_comp) ]
        self.parameter_values = {}

    def set_linear_rates(self,rate_list,reset_rates=True,allow_nonzero_column_sums=True):
        """
        Define the linear transition rates between compartments.

        Parameters
        ==========
        rate_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ( acting_compartment, affected_compartment, rate ),
                    ...
                ]

        allow_nonzero_column_sums : :obj:`bool`, default : False
            This keyword has no function in this class
        reset_rates : bool, default : True
            Whether to reset all linear rates to zero before
            setting the new ones.
        """

        if reset_rates:
            linear_rates = sympy.zeros(self.N_comp, self.N_comp)
            birth_rates = sympy.zeros(self.N_comp,1)
            self.has_functional_rates = False
        else:
            linear_rates = sympy.Matrix(self.linear_rates)
            birth_rates = sympy.Matrix(self.birth_rates)

        for acting_compartment, affected_compartment, rate in rate_list:

            _t = self.get_compartment_id(affected_compartment)
            if acting_compartment is None:
                birth_rates[_t] += rate
            else:
                _s = self.get_compartment_id(acting_compartment)
                linear_rates[_t, _s] += rate

            self._check_rate_for_functional_dependency(rate)

        self.linear_rates = linear_rates
        self.birth_rates = birth_rates

        return self

    def _check_rate_for_functional_dependency(self,rate):
        try:
            self.has_functional_rates |= any([ compartment in rate.free_symbols for compartment in self.compartments])
        except AttributeError as e:
            return


    def set_quadratic_rates(self,rate_list,reset_rates=True,allow_nonzero_column_sums=False):
        r"""
        Define the quadratic transition processes between compartments.

        Parameters
        ----------
        rate_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ("coupling_compartment_0",
                     "coupling_compartment_1",
                     "affected_compartment",
                     rate
                     ),
                    ...
                ]

        allow_nonzero_column_sums : :obj:`bool`, default : False
            This keyword has no function in this class
        reset_rates : bool, default : True
            Whether to reset all quadratic rates to zero before
            setting the new ones.

        Example
        -------
        For an SEIR model.

        .. code:: python

            epi.set_quadratic_rates([
                ("S", "I", "S", -1 ),
                ("S", "I", "E", +1 )
            ])

        Read  as

        "Coupling of *S* and *I* leads to
        a reduction in "S" proportional to :math:`S\times I` and rate -1/time_unit.
        Furthermore, coupling of *S* and *I* leads to
        an increase in "E" proportional to :math:`S\times I` and rate +1/time_unit."
        """

        if reset_rates:

            matrices = [None for c in self.compartments]
            for c in range(self.N_comp):
                matrices[c] = sympy.zeros(self.N_comp, self.N_comp)

            all_affected = []
            self.has_functional_rates = False
        else:
            matrices =  [ sympy.Matrix(M) for M in self.quadratic_rates ]
            all_affected = self.affected_by_quadratic_process if len(self.affected_by_quadratic_process)>0 else []

        for coupling0, coupling1, affected, rate in rate_list:

            c0, c1 = sorted([ self.get_compartment_id(c) for c in [coupling0, coupling1] ])
            a = self.get_compartment_id(affected)

            self._check_rate_for_functional_dependency(rate)

            matrices[a][c0,c1] += rate
            all_affected.append(a)

        self.affected_by_quadratic_process = sorted(list(set(all_affected)))
        self.quadratic_rates = matrices

        return self


    def dydt(self):
        """
        Obtain the equations of motion for this model in form of a sympy.Matrix.
        """

        y = sympy.Matrix(self.compartments)

        ynew = self.linear_rates * y + self.birth_rates
        if self.correct_for_dynamical_population_size:
            population_size = sum(self.compartments)
        else:
            population_size = self.initial_population_size
        for c in self.affected_by_quadratic_process:
            ynew[c] += (y.T * self.quadratic_rates[c] * y)[0,0] / population_size

        return ynew

class SymbolicMatrixSIModel(SymbolicMatrixEpiModel):
    """
    An SI model derived from :class:`epipack.symbolic_epi_models.SymbolicMatrixEpiModel`.
    """

    def __init__(self, infection_rate, initial_population_size=1):

        S, I = sympy.symbols("S I")

        SymbolicMatrixEpiModel.__init__(self,[S, I], initial_population_size)

        self.set_processes([
                (S, I, infection_rate, I, I),
            ])

class SymbolicMatrixSIRModel(SymbolicMatrixEpiModel):
    """
    An SIR model derived from :class:`epipack.symbolic_epi_models.SymbolicMatrixEpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, initial_population_size=1):

        S, I, R = sympy.symbols("S I R")

        SymbolicMatrixEpiModel.__init__(self,[S, I, R], initial_population_size)

        self.add_transmission_processes([
                (S, I, infection_rate, I, I),
            ])

        self.add_transition_processes([
                (I, recovery_rate, R),
            ])

class SymbolicMatrixSISModel(SymbolicMatrixEpiModel):
    """
    An SIS model derived from :class:`epipack.symbolic_epi_models.SymbolicMatrixEpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, initial_population_size=1):

        S, I = sympy.symbols("S I")

        SymbolicMatrixEpiModel.__init__(self,[S, I], initial_population_size)

        self.add_transmission_processes([
                (S, I, infection_rate, I, I),
            ])

        self.add_transition_processes([
                (I, recovery_rate, S),
            ])

class SymbolicMatrixSIRSModel(SymbolicMatrixEpiModel):
    """
    An SIRS model derived from :class:`epipack.symbolic_epi_models.SymbolicMatrixEpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, waning_immunity_rate, initial_population_size=1):

        S, I, R = sympy.symbols("S I R")

        SymbolicMatrixEpiModel.__init__(self,[S, I, R], initial_population_size)

        self.add_transmission_processes([
                (S, I, infection_rate, I, I),
            ])

        self.add_transition_processes([
                (I, recovery_rate, R),
                (R, waning_immunity_rate, S),
            ])

if __name__=="__main__":    # pragma: no cover

    import sympy

    S, I, eta, rho = sympy.symbols("S I eta rho")

    SIS = SymbolicMatrixEpiModel([S,I])
    SIS.add_transmission_processes([
            (I, S, eta, I, I),
        ])
    SIS.add_transition_processes([
            (I, rho, S),
        ])

    print(SIS.find_fixed_points())

    print(SIS.get_eigenvalues_at_fixed_point({S:1}))

    print("==========")
    SIS = SymbolicMatrixEpiModel([S,I])
    SIS.set_processes([
            (I, S, eta/(1-I), I, I),
            (I, rho, S),
        ])
    print(SIS.jacobian())
    print(SIS.get_eigenvalues_at_disease_free_state())


    print("gray scott")
    u, v, f, k = sympy.symbols("u v f k")
    GS = SymbolicMatrixEpiModel([u,v])
    GS.set_linear_rates([
            (None, u, f),
            (u, u, -f),
            (v, v, -f-k),
        ])

    GS.set_quadratic_rates([
            (u, v, u, -v),
            (u, v, v, +v),
        ])


    GS.set_processes([
            (u, f, None),
            (None, f, u),
            (v, f+k, None),
            (u, v, v*1, v, v),
        ],ignore_rate_position_checks=True)

    print(GS.ODEs())

    print(GS.find_fixed_points())

    print("===========")


    N, omega = sympy.symbols("N omega")

    epi = SymbolicMatrixSIRSModel(eta, rho, omega, initial_population_size=N)
    print()
    print(epi.ODEs())
    print(epi.find_fixed_points())

    print("==========")
    x = sympy.symbols("x")
    SIS = SymbolicMatrixEpiModel([x,I])
    SIS.set_processes([
            (I, x, eta/(1-I), I, I),
            (I, rho, x),
        ])
    try:
        print(SIS.get_eigenvalues_at_disease_free_state())
    except ValueError as e:
        print(e)

    print("===========")
    S, I = sympy.symbols("S I")
    epi = SymbolicMatrixSISModel(eta, rho)

    epi.set_initial_conditions({S: 1-0.01, I:0.01 })
    epi.set_parameter_values({eta:2,rho:1})

    tt = np.linspace(0,10,1000)
    result = epi.integrate(tt)

    import matplotlib.pyplot as pl

    pl.plot(tt, result[S])
    pl.plot(tt, result[I])

    print("===========")
    t, S, I = sympy.symbols("t S I")
    epi = SymbolicMatrixSISModel((1.5+sympy.cos(t))*eta, rho)

    epi.set_initial_conditions({S: 1-0.2, I:0.2 })
    epi.set_parameter_values({eta:2,rho:1})

    tt = np.linspace(0,20,1000)
    result = epi.integrate(tt)

    import matplotlib.pyplot as pl

    pl.figure()
    pl.plot(tt, result[S])
    pl.plot(tt, result[I])

    print("===========")
    t, S, I = sympy.symbols("t S I")
    epi = SymbolicMatrixSISModel(eta/(1-S), rho)

    epi.set_initial_conditions({S: 1-0.2, I:0.2 })
    epi.set_parameter_values({eta:2,rho:1})

    tt = np.linspace(0,5,1000)
    result = epi.integrate(tt)

    import matplotlib.pyplot as pl

    pl.figure()
    pl.plot(tt, result[S])
    pl.plot(tt, result[I])
    pl.show()




"""
Contains a MockSamplableSet class that mimicks the behavior of 
github.com/gstonge/SamplableSet but is less efficient.
"""

import numpy as np

def choice(arr,p):
    """
    Returns a random element from ``arr`` with probability given in array ``p``.
    If ``arr`` is not an iterable, the function returns the index of the chosen element.
    """
    ndx = np.argmax(np.random.rand()<np.cumsum(p))
    try:
        return arr[ndx]
    except TypeError as e:
        return ndx


class MockSamplableSet:
    """
    A set of items that can be sampled with probability
    proportional to a corresponding item weight.

    Mimicks the behavior of github.com/gstonge/SamplableSet
    without being as efficient.

    Works similar to Python's set, with ``__getitem__``, 
    ``__setitem__``, ``__delitem__``, ``__iter__``,
    ``__len__``, ``__contains__``.

    Parameters
    ==========
    min_weight : float
        minimum possible weight
    max_weight : float
        maximum possible weight
    weighted_elements : list, default = []
        list of 2-tuples, first entry an item, second entry a weight
    cpp_type : str, default = 'int'
        The type of the items.

    Attributes
    ==========
    min_weight : float
        minimum possible weight
    max_weight : float
        maximum possible weight
    items : numpy.ndarray
        list of items in this set
    weights : numpy.ndarray
        list of corresponding weights
    """

    def __init__(self,min_weight,max_weight,weighted_elements=[],cpp_type='int'):

        self.min_weight = min_weight
        self.max_weight = max_weight

        if type(weighted_elements) == dict:
            weighted_elements = list(weighted_elements.items())

        self.items = np.array([ e[0] for e in weighted_elements ],dtype=cpp_type)
        self.weights = np.array([ e[1] for e in weighted_elements ],dtype=float)
        sort_ndx = np.argsort(self.items)
        self.items = self.items[sort_ndx]
        self.weights = self.weights[sort_ndx]
        self._total_weight = self.weights.sum()

        if np.any(self.weights < self.min_weight):
            raise ValueError("There are weights below the limit.")

        if np.any(self.weights > self.max_weight):
            raise ValueError("There are weights above the limit.")

    def sample(self):
        """
        Random sample from the set, sampled
        with probability proportional to items' weight.

        Returns
        =======
        item : cpp_type
            An item from the set
        weight : float
            The weight of the item
        """

        #ndx = np.random.choice(len(self.items),p=self.weights/self._total_weight)
        #ndx = np.argwhere(np.random.rand()<np.cumsum(self.weights/self._total_weight))[0][0]
        ndx = choice(len(self.items),p=self.weights/self._total_weight)
        return self.items[ndx], self.weights[ndx]

    def __getitem__(self,key):
        found_key, ndx = self._find_key(key)
        if not found_key:
            raise KeyError("`",key,"` is not in this set.")
        else:
            return self.weights[ndx]

    def __delitem__(self,key):
        found_key, ndx = self._find_key(key)
        if found_key:
            self.items = np.delete(self.items, ndx)
            self.weights = np.delete(self.weights, ndx)
            self._total_weight = self.weights.sum()

    def __setitem__(self,key,value):
        if value < self.min_weight or value > self.max_weight:
            raise ValueError('Inserting element-weight pair ' + str(key) +" "+ str(value)+" \n" +\
                             'has weight value out of bounds of ' + str(self.min_weight) + " " + \
                             str(self.max_weight))
        found_key, ndx = self._find_key(key) 
        if not found_key:
            self.items = np.insert(self.items, ndx, key)
            self.weights = np.insert(self.weights, ndx, value)
        else:
            self.weights[ndx] = value
            
        self._total_weight = self.weights.sum()

    def _find_key(self,key):
        ndx = np.searchsorted(self.items, key)
        return ( not ((ndx == len(self.items) or self.items[ndx] != key)), ndx )

    def __iter__(self):
        self._ndx = 0
        return self

    def __next__(self):
        if self._ndx < len(self.items):
            i, w = self.items[self._ndx], self.weights[self._ndx]            
            self._ndx += 1
            return (i,w)
        else:
            raise StopIteration

    def __len__(self):
        return len(self.items)

    def __contains__(self,key):
        return self._find_key(key)[0]

    def total_weight(self):
        """Obtain the total weight of the set"""
        return self._total_weight

    def clear(self):
        """Reset the set. Not implemented yet."""
        pass



if __name__ == "__main__":    # pragma: no cover

    s = MockSamplableSet(1.0,2.0,{0:1.2,3:1.8,})

    print('===========')
    print(s.items)
    print(s.weights)
    print(len(s))
    print(s.total_weight())

    np.random.seed(1)
    print('===========')
    for _ in range(5):
        print(s.sample())


    print('===========')
    print(s[0])
    print(s[0])

    s[0] = 2
    print('===========')
    print(s.items)
    print(s.weights)
    print(s.total_weight())

    s[1] = 1.3
    print('===========')
    print(s.items)
    print(s.weights)
    print(s.total_weight())

    del s[3]
    print('===========')
    print(s.items)
    print(s.weights)
    print(s.total_weight())

    print('===========')
    for item, weight in s:
        print(item, weight)

    print('===========')
    print(0 in s)
    print(45 in s)




from setuptools import setup, Extension
import setuptools
import os
import sys

# get __version__, __author__, and __email__
exec(open("./epipack/metadata.py").read())

setup(
    name='epipack',
    version=__version__,
    author=__author__,
    author_email=__email__,
    url='https://github.com/benmaier/epipack',
    license=__license__,
    description="Fast prototyping of epidemiological models based on reaction equations. Analyze the ODEs analytically or numerically, or run stochastic simulations on networks/well-mixed systems.",
    long_description='',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
                'numpy>=1.17',
                'scipy>=1.3',
                'sympy>=1.6',
                'pyglet>=1.5.15,<1.6',
                'ipython>=7.14.0',
                'ipywidgets>=7.5.1',
                'matplotlib>=3.0.0',
    ],
    tests_require=['pytest', 'pytest-cov'],
    setup_requires=['pytest-runner'],
    classifiers=['License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 ],
    project_urls={
        'Documentation': 'http://epipack.benmaier.org',
        'Contributing Statement': 'https://github.com/benmaier/epipack/blob/master/CONTRIBUTING.md',
        'Bug Reports': 'https://github.com/benmaier/epipack/issues',
        'Source': 'https://github.com/benmaier/epipack/',
        'PyPI': 'https://pypi.org/project/epipack/',
    },
    include_package_data=True,
    zip_safe=False,
)
